"""
Tool for merging pod5 files
"""

from time import sleep
from typing import Iterable
from pathlib import Path
from tqdm.auto import tqdm

import pod5 as p5
import pod5.repack as p5_repack
from pod5.tools.parsers import prepare_pod5_merge_argparser, run_tool
from pod5.tools.utils import (
    DEFAULT_THREADS,
    PBAR_DEFAULTS,
    collect_inputs,
    init_logging,
    logged_all,
)

logger = init_logging()


@logged_all
def merge_pod5(
    inputs: Iterable[Path],
    output: Path,
    duplicate_ok: bool = False,
    force_overwrite: bool = False,
    recursive: bool = False,
    threads: int = DEFAULT_THREADS,
    readers: int = 5,
) -> None:
    """
    Merge the an iterable of input pod5 paths into the specified output path
    """

    if output.exists():
        if force_overwrite:
            output.unlink()
        else:
            raise FileExistsError(
                f"Output files already exists and --force-overwrite not set. "
                f"Refusing to overwrite {output}."
            )

    if not output.parent.exists():
        output.parent.mkdir(parents=True, exist_ok=True)

    _inputs = collect_inputs(
        inputs, recursive=recursive, pattern="*.pod5", threads=threads
    )

    print(f"Merging reads from {len(_inputs)} files")

    # Open the output file writer
    with p5.Writer(output.absolute()) as writer:
        # Attach the writer to the repacker
        repacker = p5_repack.Repacker()
        repacker_output = repacker.add_output(writer, not duplicate_ok)

        pbar = tqdm(
            total=len(_inputs),
            desc="Merging",
            unit="Files",
            leave=True,
            position=0,
            **PBAR_DEFAULTS,
        )

        active_limit = max(readers, 1)
        opened_readers = 0
        active = 0

        while _inputs or active > 0:
            pbar.update(opened_readers - active - pbar.n)

            active = repacker.currently_open_file_reader_count
            if active < active_limit:
                if _inputs:
                    with p5.Reader(_inputs.pop()) as reader:
                        opened_readers += 1
                        repacker.add_all_reads_to_output(repacker_output, reader)
                        continue
                else:
                    repacker.set_output_finished(repacker_output)
                    continue

            sleep(0.05)

        repacker.finish()
        del repacker
        pbar.update(opened_readers - active - pbar.n)
        pbar.close()

    return


def main():
    """pod5_merge main program"""
    run_tool(prepare_pod5_merge_argparser())


if __name__ == "__main__":
    main()



"""
Tools for accessing POD5 data from PyArrow files
"""


import mmap
from collections import namedtuple
from dataclasses import fields
from io import IOBase
import os
from pathlib import Path
from typing import (
    Collection,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)
from uuid import UUID

import lib_pod5 as p5b
import numpy as np
import numpy.typing as npt
import packaging.version
import pyarrow as pa

from pod5.pod5_types import (
    Calibration,
    EndReason,
    EndReasonEnum,
    PathOrStr,
    Pore,
    Read,
    RunInfo,
    ShiftScalePair,
)

from .api_utils import Pod5ApiException, format_read_ids, pack_read_ids, safe_close
from .signal_tools import vbz_decompress_signal, vbz_decompress_signal_into


ReadRecordV3Columns = namedtuple(
    "ReadRecordV3Columns",
    [
        "read_id",
        "read_number",
        "start",
        "channel",
        "well",
        "median_before",
        "pore_type",
        "calibration_offset",
        "calibration_scale",
        "end_reason",
        "end_reason_forced",
        "run_info",
        "signal",
        "num_minknow_events",
        "tracked_scaling_scale",
        "tracked_scaling_shift",
        "predicted_scaling_scale",
        "predicted_scaling_shift",
        "num_reads_since_mux_change",
        "time_since_mux_change",
        "num_samples",
    ],
)


Signal = namedtuple("Signal", ["signal", "samples"])
SignalRowInfo = namedtuple(
    "SignalRowInfo",
    ["batch_index", "batch_row_index", "sample_count", "byte_count"],
)


class ReadRecord:
    """
    Represents the data for a single read from a pod5 record.
    """

    def __init__(
        self,
        reader: "Reader",
        batch: "ReadRecordBatch",
        row: int,
        batch_signal_cache: Optional[List[npt.NDArray[np.int16]]] = None,
        selected_batch_index: Optional[int] = None,
    ):
        """ """
        self._reader = reader
        self._batch = batch
        self._row = row
        self._batch_signal_cache = batch_signal_cache
        self._selected_batch_index = selected_batch_index

    @property
    def read_id(self) -> UUID:
        """
        Get the unique read identifier for the read as a `UUID`.
        """
        return UUID(bytes=self._batch.columns.read_id[self._row].as_py())

    @property
    def read_number(self) -> int:
        """
        Get the integer read number of the read.
        """
        return self._batch.columns.read_number[self._row].as_py()  # type: ignore

    @property
    def start_sample(self) -> int:
        """
        Get the absolute sample which the read started.
        """
        return self._batch.columns.start[self._row].as_py()  # type: ignore

    @property
    def num_samples(self) -> int:
        """
        Get the number of samples in the reads signal data.
        """
        return self._batch.columns.num_samples[self._row].as_py()  # type: ignore

    @property
    def median_before(self) -> float:
        """
        Get the median before level (in pico amps) for the read.
        """
        return self._batch.columns.median_before[self._row].as_py()  # type: ignore

    @property
    def num_minknow_events(self) -> int:
        """
        Find the number of minknow events in the read.
        """
        return self._batch.columns.num_minknow_events[self._row].as_py()  # type: ignore

    @property
    def tracked_scaling(self) -> ShiftScalePair:
        """
        Find the tracked scaling value in the read.
        """
        return ShiftScalePair(
            self._batch.columns.tracked_scaling_shift[self._row].as_py(),
            self._batch.columns.tracked_scaling_scale[self._row].as_py(),
        )

    @property
    def predicted_scaling(self) -> ShiftScalePair:
        """
        Find the predicted scaling value in the read.
        """
        return ShiftScalePair(
            self._batch.columns.predicted_scaling_shift[self._row].as_py(),
            self._batch.columns.predicted_scaling_scale[self._row].as_py(),
        )

    @property
    def num_reads_since_mux_change(self) -> int:
        """
        Number of selected reads since the last mux change on this reads channel.
        """
        return self._batch.columns.num_reads_since_mux_change[self._row].as_py()  # type: ignore

    @property
    def time_since_mux_change(self) -> float:
        """
        Time in seconds since the last mux change on this reads channel.
        """
        return self._batch.columns.time_since_mux_change[self._row].as_py()  # type: ignore

    @property
    def pore(self) -> Pore:
        """
        Get the pore data associated with the read.
        """
        return Pore(
            self._batch.columns.channel[self._row].as_py(),
            self._batch.columns.well[self._row].as_py(),
            self._batch.columns.pore_type[self._row].as_py(),
        )

    @property
    def calibration(self) -> Calibration:
        """
        Get the calibration data associated with the read.
        """
        return Calibration(
            self._batch.columns.calibration_offset[self._row].as_py(),
            self._batch.columns.calibration_scale[self._row].as_py(),
        )

    @property
    def calibration_digitisation(self) -> int:
        """
        Get the digitisation value used by the sequencer.

        Intended to assist workflows ported from legacy file formats.
        """
        return self.run_info.adc_max - self.run_info.adc_min + 1

    @property
    def calibration_range(self) -> float:
        """
        Get the calibration range value.

        Intended to assist workflows ported from legacy file formats.
        """
        return self.calibration.scale * self.calibration_digitisation

    @property
    def end_reason(self) -> EndReason:
        """
        Get the end reason data associated with the read.
        """
        return EndReason(
            reason=EndReasonEnum[
                self._batch.columns.end_reason[self._row].as_py().upper()
            ],
            forced=self._batch.columns.end_reason_forced[self._row].as_py(),
        )

    @property
    def run_info(self) -> RunInfo:
        """
        Get the run info data associated with the read.
        """
        return self._reader._lookup_run_info(self._batch, self._row)

    @property
    def end_reason_index(self) -> int:
        """
        Get the dictionary index of the end reason data associated with the read.
        This property is the same as the EndReason enumeration value.
        """
        return self._batch.columns.end_reason[self._row].index.as_py()  # type: ignore

    @property
    def run_info_index(self) -> int:
        """
        Get the dictionary index of the run info data associated with the read.
        """
        return self._batch.columns.run_info[self._row].index.as_py()  # type: ignore

    @property
    def sample_count(self) -> int:
        """
        Get the number of samples in the reads signal data.
        """
        return self.num_samples

    @property
    def byte_count(self) -> int:
        """
        Get the number of bytes used to store the reads data.
        """
        return sum(r.byte_count for r in self.signal_rows)

    @property
    def has_cached_signal(self) -> bool:
        """
        Get if cached signal is available for this read.
        """
        return self._batch_signal_cache is not None

    @property
    def signal(self) -> npt.NDArray[np.int16]:
        """
        Get the full signal for the read.

        Returns
        -------
        numpy.ndarray[int16]
            A numpy array of signal data with int16 type.
        """
        if self._batch_signal_cache is not None:
            if self._selected_batch_index is not None:
                return self._batch_signal_cache[self._selected_batch_index]
            return self._batch_signal_cache[self._row]

        rows = self._batch.columns.signal[self._row]
        batch_data = [self._find_signal_row_index(r.as_py()) for r in rows]
        sample_counts = []
        for batch, _, batch_row_index in batch_data:
            sample_counts.append(batch.samples[batch_row_index].as_py())

        output = np.empty(dtype=np.int16, shape=(sum(sample_counts),))
        current_sample_index = 0

        for i, (batch, _, batch_row_index) in enumerate(batch_data):
            signal = batch.signal
            current_row_count = sample_counts[i]
            output_slice = output[
                current_sample_index : current_sample_index + current_row_count
            ]
            if self._reader.is_vbz_compressed:
                vbz_decompress_signal_into(
                    memoryview(signal[batch_row_index].as_buffer()), output_slice
                )
            else:
                output_slice[:] = signal.to_numpy()
            current_sample_index += current_row_count
        return output

    @property
    def signal_pa(self) -> npt.NDArray[np.float32]:
        """
        Get the full signal for the read, calibrated in pico amps.

        Returns
        -------
        numpy.ndarray[float32]
            A numpy array of signal data in pico amps with float32 type.
        """
        return self.calibrate_signal_array(self.signal)

    def signal_for_chunk(self, index: int) -> npt.NDArray[np.int16]:
        """
        Get the signal for a given chunk of the read.

        Returns
        -------
        numpy.ndarray[int16]
            A numpy array of signal data with int16 type for the specified chunk.
        """
        # signal_rows can be used to find details of the signal chunks.
        chunk_abs_row_index = self._batch.columns.signal[self._row][index]
        return self._get_signal_for_row(chunk_abs_row_index.as_py())

    @property
    def signal_rows(self) -> List[SignalRowInfo]:
        """
        Get all signal rows for the read

        Returns
        -------
        list[SignalRowInfo]
            A list of signal row data (as SignalRowInfo) in the read.
        """

        def map_signal_row(sig_row) -> SignalRowInfo:
            sig_row = sig_row.as_py()

            batch, batch_index, batch_row_index = self._find_signal_row_index(sig_row)
            return SignalRowInfo(
                batch_index,
                batch_row_index,
                batch.samples[batch_row_index].as_py(),
                len(batch.signal[batch_row_index].as_buffer()),
            )

        return [map_signal_row(r) for r in self._batch.columns.signal[self._row]]

    def calibrate_signal_array(
        self, signal_array_adc: npt.NDArray[np.int16]
    ) -> npt.NDArray[np.float32]:
        """
        Transform an array of int16 signal data from ADC space to pA.

        Returns
        -------
        A numpy array of signal data with float32 type.
        """
        offset = np.float32(self.calibration.offset)
        scale = np.float32(self.calibration.scale)
        return (signal_array_adc + offset) * scale

    def _find_signal_row_index(self, signal_row: int) -> Tuple[Signal, int, int]:
        """
        Map from a signal_row to a Signal, batch index and row index within that batch.

        Returns
        -------
        A Tuple containing the `Signal` and its `batch_index` and `row_index`
        """
        sig_row_count: int = self._reader.signal_batch_row_count
        sig_batch_idx: int = signal_row // sig_row_count
        sig_batch = self._reader._get_signal_batch(sig_batch_idx)
        batch_row_idx: int = signal_row - (sig_batch_idx * sig_row_count)

        return sig_batch, sig_batch_idx, batch_row_idx

    def _get_signal_for_row(self, signal_row: int) -> npt.NDArray[np.int16]:
        """
        Get the signal data for a given absolute signal row index

        Returns
        -------
        A numpy array of signal data with int16 type.
        """
        batch, _, batch_row_index = self._find_signal_row_index(signal_row)

        signal = batch.signal
        if self._reader.is_vbz_compressed:
            sample_count = batch.samples[batch_row_index].as_py()
            return vbz_decompress_signal(
                memoryview(signal[batch_row_index].as_buffer()), sample_count
            )

        return signal.to_numpy()

    def to_read(self) -> Read:
        """
        Create a mutable :py:class:`pod5.pod5_types.Read` from this
        :py:class:`ReadRecord` instance.

        Returns
        -------
            :py:class:`pod5.pod5_types.Read`
        """
        return Read(
            read_id=self.read_id,
            pore=self.pore,
            calibration=self.calibration,
            median_before=self.median_before,
            end_reason=self.end_reason,
            read_number=self.read_number,
            run_info=self.run_info,
            start_sample=self.start_sample,
            signal=self.signal,
        )


class ReadRecordBatch:
    """
    Read data for a batch of reads.
    """

    def __init__(self, reader: "Reader", batch: pa.RecordBatch):
        """ """

        self._reader: "Reader" = reader
        self._batch: pa.RecordBatch = batch

        self._signal_cache: Optional[p5b.Pod5SignalCacheBatch] = None
        self._selected_batch_rows: Optional[Iterable[int]] = None
        self._columns: Optional[ReadRecordV3Columns] = None

    @property
    def columns(self) -> ReadRecordV3Columns:
        """Return the data from this batch as a ReadRecordColumns instance"""
        if self._columns is None:
            self._columns = ReadRecordV3Columns(
                *[
                    self._batch.column(name)
                    for name in self._reader._columns_type._fields
                ]
            )
        return self._columns

    def set_cached_signal(self, signal_cache: p5b.Pod5SignalCacheBatch) -> None:
        """Set the signal cache"""
        self._signal_cache = signal_cache

    def set_selected_batch_rows(self, selected_batch_rows: Iterable[int]) -> None:
        """Set the selected batch rows"""
        self._selected_batch_rows = selected_batch_rows

    def reads(self) -> Generator[ReadRecord, None, None]:
        """
        Iterate all reads in this batch.

        Yields
        ------
        ReadRecord
            ReadRecord instances in the file.
        """

        signal_cache = None
        if self._signal_cache and self._signal_cache.samples:
            signal_cache = self._signal_cache.samples

        if self._selected_batch_rows is not None:
            for idx, row in enumerate(self._selected_batch_rows):
                yield ReadRecord(
                    self._reader,
                    self,
                    row,
                    batch_signal_cache=signal_cache,
                    selected_batch_index=idx,
                )
        else:
            for i in range(self.num_reads):
                yield ReadRecord(self._reader, self, i, batch_signal_cache=signal_cache)

    def get_read(self, row: int) -> ReadRecord:
        """Get the ReadRecord at row index"""
        return ReadRecord(self._reader, self, row)

    @property
    def num_reads(self) -> int:
        """Return the number of rows in this RecordBatch"""
        return int(self._batch.num_rows)

    @property
    def read_id_column(self) -> pa.FixedSizeBinaryArray:
        """
        Get the column of read ids for this batch
        """
        if self._selected_batch_rows is not None:
            return self.columns.read_id.take(self._selected_batch_rows)
        return self.columns.read_id

    @property
    def read_number_column(self) -> pa.UInt32Array:
        """
        Get the column of read numbers for this batch
        """
        if self._selected_batch_rows is not None:
            return self.columns.read_number.take(self._selected_batch_rows)
        return self.columns.read_number

    @property
    def cached_sample_count_column(self) -> npt.NDArray[np.uint64]:
        """
        Get the sample counts from the cached signal data
        """
        if not self._signal_cache:
            raise RuntimeError("No cached signal data available")
        return self._signal_cache.sample_count

    @property
    def cached_samples_column(self) -> List[npt.NDArray[np.int16]]:
        """
        Get the samples column from the cached signal data
        """
        if not self._signal_cache:
            raise RuntimeError("No cached signal data available")
        return self._signal_cache.samples


class ArrowTableHandle:
    """Class for managing arrow file handles and memory view mapping of tables"""

    def __init__(self, location: p5b.EmbeddedFileData) -> None:
        """
        Open a pod5 file at the given `path` and use the location data to load
        an arrow table (e.g. signal table)

        Parameters
        ----------
        location : lib_pod5.pod5_format_pybind.EmbeddedFileData
            Location data for how a pod5 file should be spit in memory to read a table.
            This is returned from p5b.Pod5FileReader.get_file_X_location methods

        Raises
        ------
        Pod5ApiException
            If handle could not be opened
        """

        # The location data is passed from the p5b.Pod5FileReader.get_file_X_location
        # methods
        self._location = location
        self._path = Path(self._location.file_path)

        # Open the file
        self._fh = self._path.open("rb")

        if "POD5_DISABLE_MMAP_OPEN" in os.environ:
            self._reader = self._open_without_mmap()
        else:
            # Create a memory view of the file and select the region for the table
            try:
                self._reader = self._open_with_mmap()
            except OSError:
                # If we fail fall back to a traditional open.
                self._reader = self._open_without_mmap()

    def _open_without_mmap(self):
        class File(IOBase):
            def __init__(self, handle, location):
                self._handle = handle
                self._location = location
                self.seek(0, whence=0)

            def seek(self, position, whence=0):
                if whence == 0:
                    position = position + self._location.offset
                elif whence == 2:
                    position = (
                        self._location.offset + self._location.length
                    ) - position
                    whence = 0
                # The new abs location:
                abs_location = self._handle.seek(position, whence)

                return abs_location - self._location.offset

            def read(self, size=-1):
                return self._handle.read(size)

        return pa.ipc.open_file(pa.PythonFile(File(self._fh, self._location)))

    def _open_with_mmap(self):
        _mmap = mmap.mmap(self._fh.fileno(), length=0, access=mmap.ACCESS_READ)
        file_view = memoryview(_mmap)

        arrow_table_view = file_view[
            self._location.offset : self._location.offset + self._location.length
        ]

        # Open the table
        try:
            return pa.ipc.open_file(pa.BufferReader(arrow_table_view))
        except pa.ArrowInvalid as exc:
            raise Pod5ApiException(f"Failed to open ArrowTable: {self._path}") from exc

    @property
    def reader(self) -> pa.ipc.RecordBatchFileReader:
        """Return the pyarrow file reader object"""
        if self._reader is not None:
            return self._reader

        raise RuntimeError(f"Could not open pyarrow reader: {p5b.get_error_string()}")

    def close(self) -> None:
        """
        Cleanly close the open file handles and memory views.
        """
        self._reader = None
        safe_close(self, "_fh")

    def __enter__(self) -> "ArrowTableHandle":
        return self

    def __exit__(self, *exc_details) -> None:
        self.close()

    def __del__(self) -> None:
        self.close()


class Reader:
    """
    The base reader for POD5 data
    """

    def __init__(self, path: PathOrStr):
        """
        Open a pod5 filepath for reading
        """

        self._path = Path(path).absolute()

        self._file_reader: Optional[p5b.Pod5FileReader] = None
        self._read_handle: Optional[ArrowTableHandle] = None
        self._run_info_handle: Optional[ArrowTableHandle] = None
        self._signal_handle: Optional[ArrowTableHandle] = None

        (
            self._file_reader,
            self._read_handle,
            self._run_info_handle,
            self._signal_handle,
        ) = self._open_arrow_table_handles(self._path)

        schema_metadata = self.read_table.schema.metadata
        self._file_identifier = UUID(
            schema_metadata[b"MINKNOW:file_identifier"].decode("utf-8")
        )
        self._writing_software = schema_metadata[b"MINKNOW:software"].decode("utf-8")
        writing_version_str = schema_metadata[b"MINKNOW:pod5_version"].decode("utf-8")
        writing_version = packaging.version.parse(writing_version_str)

        self._columns_type = ReadRecordV3Columns
        self._reads_table_version = 3

        self._file_version = writing_version
        self._file_version_pre_migration = packaging.version.Version(
            self._file_reader.get_file_version_pre_migration()
        )

        # Warning: The cached signal maintains an open file handle. So ensure that
        # this dictionary is cleared before closing.
        self._cached_signal_batches: Dict[int, Signal] = {}
        self._cached_run_infos: Dict[str, RunInfo] = {}

        self._is_vbz_compressed: Optional[bool] = None
        self._signal_batch_row_count: Optional[int] = None

    @staticmethod
    def _open_arrow_table_handles(
        path: Path,
    ) -> Tuple[
        p5b.Pod5FileReader, ArrowTableHandle, ArrowTableHandle, ArrowTableHandle
    ]:
        """Open handles to the underlying arrow tables within this pod5 file"""
        if not path.is_file():
            raise FileNotFoundError(f"Failed to open pod5 file at: {path}")

        file_reader = p5b.open_file(str(path))
        if not file_reader:
            raise Pod5ApiException(
                f"Failed to open reader for {path} Reason: {p5b.get_error_string()}"
            )

        read_handle = ArrowTableHandle(file_reader.get_file_read_table_location())
        run_info_handle = ArrowTableHandle(
            file_reader.get_file_run_info_table_location()
        )
        signal_handle = ArrowTableHandle(file_reader.get_file_signal_table_location())

        return file_reader, read_handle, run_info_handle, signal_handle

    def __del__(self) -> None:
        self.close()

    def __enter__(self) -> "Reader":
        return self

    def __exit__(self, *exc_details) -> None:
        self.close()

    def __iter__(self) -> Generator[ReadRecord, None, None]:
        """Iterate over all reads"""
        yield from self.reads()

    def close(self) -> None:
        """Close files handles"""

        safe_close(self, "_read_handle")
        self._read_handle = None

        safe_close(self, "_run_info_handle")
        self._run_info_handle = None

        safe_close(self, "_signal_handle")
        self._signal_handle = None

        safe_close(self, "_file_reader")
        self._file_reader = None

        # Explicitly clear this dictionary to close file handles used in cache
        self._cached_signal_batches = {}

    @property
    def path(self) -> Path:
        """Return the path to this pod5 file"""
        return self._path

    @property
    def inner_file_reader(self) -> p5b.Pod5FileReader:
        """Access the inner c_api Pod5FileReader - use with caution"""
        if self._file_reader is None:
            raise RuntimeError("Pod5FileReader has been closed!")
        return self._file_reader

    @property
    def read_table(self) -> pa.ipc.RecordBatchFileReader:
        """Access the pod5 read table"""
        if self._read_handle is None:
            raise RuntimeError("ArrowTableHandle has been closed!")
        return self._read_handle.reader

    @property
    def run_info_table(self) -> pa.ipc.RecordBatchFileReader:
        """Access the pod5 run_info table"""
        if self._run_info_handle is None:
            raise RuntimeError("ArrowTableHandle has been closed!")
        return self._run_info_handle.reader

    @property
    def signal_table(self) -> pa.ipc.RecordBatchFileReader:
        """Access the pod5 signal table - use with caution"""
        if self._signal_handle is None:
            raise RuntimeError("ArrowTableHandle has been closed!")
        return self._signal_handle.reader

    @property
    def file_version(self) -> packaging.version.Version:
        return self._file_version

    @property
    def file_version_pre_migration(self) -> packaging.version.Version:
        return self._file_version_pre_migration

    @property
    def writing_software(self) -> str:
        return self._writing_software

    @property
    def file_identifier(self) -> UUID:
        return self._file_identifier

    @property
    def reads_table_version(self) -> int:
        return self._reads_table_version

    @property
    def is_vbz_compressed(self) -> bool:
        """Return if this file's signal is compressed"""
        if self._is_vbz_compressed is None:
            self._is_vbz_compressed = self.signal_table.schema.field(
                "signal"
            ).type.equals(pa.large_binary())
        return self._is_vbz_compressed

    @property
    def signal_batch_row_count(self) -> int:
        """Return signal batch row count"""
        if self._signal_batch_row_count is None:
            if self.signal_table.num_record_batches > 0:
                self._signal_batch_row_count = self.signal_table.get_batch(0).num_rows
            else:
                self._signal_batch_row_count = 0
        return self._signal_batch_row_count

    @property
    def batch_count(self) -> int:
        """
        Find the number of read batches available in the file.
        """
        return self.read_table.num_record_batches

    @property
    def num_reads(self) -> int:
        """
        Find the number of reads in the file.
        """
        return sum(batch.num_reads for batch in self.read_batches())

    @property
    def read_ids_raw(self) -> pa.ChunkedArray:
        """
        Return chunked arrow array of read ids.

        To get read ids as string use `Reader.read_ids`
        """

        return pa.chunked_array([batch.read_id_column for batch in self.read_batches()])

    @property
    def read_ids(self) -> List[str]:
        """
        Return all read_ids as a list of strings.

        For the most performant implementation consider `Reader.read_ids_raw`
        """

        def arrow_to_numpy(batch):
            # Get the arrow data as a buffer
            id_buffer = batch.read_id_column.buffers()[1]

            # Pack the arrow buffer into a numpy array of the the right shape
            array = np.frombuffer(id_buffer, dtype=np.uint8)
            return array.reshape((batch.num_reads, 16))

        read_ids = np.concatenate(
            [arrow_to_numpy(batch) for batch in self.read_batches()]
        )
        return format_read_ids(read_ids)

    def get_batch(self, index: int) -> ReadRecordBatch:
        """
        Get a read batch in the file.

        Returns
        -------
        :py:class:`ReadRecordBatch`
            The requested batch as a ReadRecordBatch.
        """
        return ReadRecordBatch(self, self.read_table.get_batch(index))

    def read_batches(
        self,
        selection: Optional[List[str]] = None,
        batch_selection: Optional[Iterable[int]] = None,
        missing_ok: bool = False,
        preload: Optional[Set[str]] = None,
    ) -> Generator[ReadRecordBatch, None, None]:
        """
        Iterate batches in the file, optionally selecting certain rows.

        Parameters
        ----------
        selection : iterable[str]
            The read ids to walk in the file.
        batch_selection : iterable[int]
            The read batches to walk in the file.
        missing_ok : bool
            If selection contains entries not found in the file, an error will be raised.
        preload : set[str]
            Columns to preload - "samples" and "sample_count" are valid values

        Returns
        -------
        An iterable of :py:class:`ReadRecordBatch` in the file.
        """
        if selection is not None:
            if batch_selection is not None:
                raise ValueError("selection and batch_selection are mutually exclusive")
            yield from self._select_read_batches(
                selection, missing_ok=missing_ok, preload=preload
            )
        elif batch_selection is not None:
            assert not selection
            yield from self._read_some_batches(batch_selection, preload=preload)
        else:
            yield from self._reads_batches(preload=preload)

    def reads(
        self,
        selection: Optional[Iterable[str]] = None,
        missing_ok: bool = False,
        preload: Optional[Set[str]] = None,
    ) -> Generator[ReadRecord, None, None]:
        """
        Iterate reads in the file, optionally filtering for certain read ids.

        Parameters
        ----------
        selection : iterable[str]
            The read ids to walk in the file.
        missing_ok : bool
            If selection contains entries not found in the file, an error will be raised.
        preload : set[str]
            Columns to preload - "samples" and "sample_count" are valid values

        Returns
        -------
        An iterable of :py:class:`ReadRecord` in the file.
        """
        if selection is None:
            yield from self._reads(preload=preload)
        else:
            yield from self._select_reads(
                list(selection), missing_ok=missing_ok, preload=preload
            )

    def _reads(
        self, preload: Optional[Set[str]] = None
    ) -> Generator[ReadRecord, None, None]:
        """Generate all reads"""
        for batch in self.read_batches(preload=preload):
            for read in batch.reads():
                yield read

    def _select_reads(
        self,
        selection: List[str],
        missing_ok: bool = False,
        preload: Optional[Set[str]] = None,
    ) -> Generator[ReadRecord, None, None]:
        """Generate selected reads"""
        for batch in self._select_read_batches(selection, missing_ok, preload=preload):
            for read in batch.reads():
                yield read

    def _reads_batches(
        self, preload: Optional[Set[str]] = None
    ) -> Generator[ReadRecordBatch, None, None]:
        """Generate the record batches"""
        signal_cache = None
        if preload:
            signal_cache = self.inner_file_reader.batch_get_signal(
                "samples" in preload,
                "sample_count" in preload,
            )

        for idx in range(self.read_table.num_record_batches):
            batch = self.get_batch(idx)
            if signal_cache:
                batch.set_cached_signal(signal_cache.release_next_batch())
            yield batch

    def _read_some_batches(
        self,
        batch_selection: Iterable[int],
        preload: Optional[Set[str]] = None,
    ) -> Generator[ReadRecordBatch, None, None]:
        """Generate the selected record batches"""
        signal_cache = None
        if preload:
            signal_cache = self.inner_file_reader.batch_get_signal_batches(
                "samples" in preload,
                "sample_count" in preload,
                np.array(batch_selection, dtype=np.uint32),
            )

        for i in batch_selection:
            batch = self.get_batch(i)
            if signal_cache:
                batch.set_cached_signal(signal_cache.release_next_batch())
            yield batch

    def _select_read_batches(
        self,
        selection: List[str],
        missing_ok: bool = False,
        preload: Optional[Set[str]] = None,
    ) -> Generator[ReadRecordBatch, None, None]:
        """Generate the selected record batches"""
        successful_finds, per_batch_counts, batch_rows = self._plan_traversal(
            selection, missing_ok=missing_ok
        )

        if not missing_ok and successful_finds != len(selection):
            raise RuntimeError(
                f"Failed to find {len(selection) - successful_finds} requested reads in the file"
            )

        signal_cache: Optional[p5b.Pod5AsyncSignalLoader] = None
        if preload:
            signal_cache = self.inner_file_reader.batch_get_signal_selection(
                "samples" in preload,
                "sample_count" in preload,
                per_batch_counts,
                batch_rows,
            )

        current_offset = 0
        for batch_idx, batch_count in enumerate(per_batch_counts):
            current_batch_rows = batch_rows[
                current_offset : current_offset + batch_count
            ]
            current_offset += batch_count

            batch = self.get_batch(batch_idx)
            batch.set_selected_batch_rows(current_batch_rows)
            if signal_cache:
                batch.set_cached_signal(signal_cache.release_next_batch())
            yield batch

    def _plan_traversal(
        self,
        read_ids: Union[Collection[str], npt.NDArray[np.uint8]],
        missing_ok: bool = False,
    ) -> Tuple[int, npt.NDArray[np.uint32], npt.NDArray[np.uint32]]:
        """
        Query the file reader indexes to return the number of read_ids which
        were found and the batches and rows which are needed to traverse each
        read in the selection.

        Parameters
        ----------
        read_ids : Collection or numpy.ndarray of read_id strings
            The read ids to find in the file

        Returns
        -------
        successful_find_count: int
            The number of reads that were found from the array of read_ids given
        per_batch_counts: numpy.array[uint32]
            The number of rows from the batch row ids to take to form each RecordBatch
        batch_rows: numpy.array[uint32]
            All batch row ids

        """
        if not isinstance(read_ids, np.ndarray):
            read_ids = pack_read_ids(read_ids, invalid_ok=missing_ok)

        assert isinstance(read_ids, np.ndarray)

        batch_rows = np.empty(dtype="u4", shape=read_ids.shape[0])
        per_batch_counts = np.empty(dtype="u4", shape=self.batch_count)

        successful_find_count = self.inner_file_reader.plan_traversal(
            read_ids,
            per_batch_counts,
            batch_rows,
        )

        return successful_find_count, per_batch_counts, batch_rows

    def _get_signal_batch(self, batch_id: int) -> Signal:
        """Get the :py:class:`Signal` from the signal_reader batch at batch_id"""
        if batch_id in self._cached_signal_batches:
            return self._cached_signal_batches[batch_id]

        batch = self.signal_table.get_batch(batch_id)

        signal_batch = Signal(*[batch.column(name) for name in Signal._fields])

        self._cached_signal_batches[batch_id] = signal_batch
        return signal_batch

    def _lookup_run_info(self, batch: ReadRecordBatch, batch_row_id: int) -> RunInfo:
        """Get the :py:class:`RunInfo` from the batch at batch_row_id"""

        acquisition_id = batch.columns.run_info[batch_row_id].as_py()

        if acquisition_id in self._cached_run_infos:
            return self._cached_run_infos[acquisition_id]

        run_info = None
        for idx in range(self.run_info_table.num_record_batches):
            run_info_batch = self.run_info_table.get_batch(idx)
            acquisition_id_col = run_info_batch.column("acquisition_id")
            for row in range(run_info_batch.num_rows):
                if acquisition_id_col[row].as_py() == acquisition_id:
                    values = {}
                    for field in fields(RunInfo):
                        col = run_info_batch.column(field.name)
                        values[field.name] = col[row].as_py()

                        if field.name in ("tracking_id", "context_tags"):
                            values[field.name] = {k: v for k, v in values[field.name]}

                    run_info = RunInfo(**values)
                    break

        if not run_info:
            raise Exception(
                f"Failed to find run info '{acquisition_id}' in run info table"
            )

        self._cached_run_infos[acquisition_id] = run_info
        return run_info



from typing import Optional
import polars as pl

# Reserved column names used in polars dataframes
PL_DEST_FNAME = "__dest_fname"
PL_SRC_FNAME = "__src_fname"
PL_READ_ID = "__read_id"
PL_UUID_REGEX = "^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"


def pl_format_read_id(read_id_col: pl.Expr) -> pl.Expr:
    """Format read ids to in UUID style"""
    read_id = read_id_col.bin.encode("hex")
    return pl.format(
        "{}-{}-{}-{}-{}",
        read_id.str.slice(0, 8),
        read_id.str.slice(8, 4),
        read_id.str.slice(12, 4),
        read_id.str.slice(16, 4),
        read_id.str.slice(20, 12),
    )


def pl_format_empty_string(expr: pl.Expr, subst: Optional[str]) -> pl.Expr:
    """Empty strings are read as a pair of double-quotes which need to be removed"""
    return pl.when(expr.str.lengths() == 0).then(subst).otherwise(expr)



"""
Tool for repacking pod5 files to potentially improve performance
"""
from concurrent.futures import ProcessPoolExecutor, as_completed
import typing
from pathlib import Path
from tqdm.auto import tqdm

import pod5 as p5
import pod5.repack
from pod5.tools.utils import (
    DEFAULT_THREADS,
    PBAR_DEFAULTS,
    assert_no_duplicate_filenames,
    collect_inputs,
    limit_threads,
)
from pod5.tools.parsers import prepare_pod5_repack_argparser, run_tool


def resolve_overwrite(src: Path, dest: Path, force: bool) -> None:
    if dest.exists():
        if dest == src:
            raise FileExistsError(f"Refusing to overwrite {src} inplace")
        if force:
            dest.unlink()
        else:
            raise FileExistsError(
                "Refusing to overwrite output without --force-overwrite"
            )


def repack_pod5_file(src: Path, dest: Path):
    """Repack the source pod5 file into dest"""
    repacker = pod5.repack.Repacker()
    with p5.Writer(dest) as writer:
        repacker_output = repacker.add_output(writer, False)
        with p5.Reader(src) as reader:
            # Add all reads to the repacker
            repacker.add_all_reads_to_output(repacker_output, reader)
        repacker.set_output_finished(repacker_output)
        repacker.finish()


def repack_pod5(
    inputs: typing.List[Path],
    output: Path,
    threads: int = DEFAULT_THREADS,
    force_overwrite: bool = False,
    recursive: bool = False,
):
    """Given a list of pod5 files, repack their contents and write files 1-1"""

    if output.exists() and not output.is_dir():
        raise ValueError(f"Output cannot be an existing file: {output}")

    # Create output directory if required
    if not output.is_dir():
        output.mkdir(parents=True, exist_ok=True)

    threads = limit_threads(threads)

    _inputs = collect_inputs(
        inputs, recursive=recursive, pattern="*.pod5", threads=threads
    )
    assert_no_duplicate_filenames(_inputs)

    # Remove existing files if required
    for input_filename in _inputs:
        output_filename = output / input_filename.name
        resolve_overwrite(input_filename, output_filename, force_overwrite)

    futures = {}
    with ProcessPoolExecutor(max_workers=threads) as executor:
        pbar = tqdm(total=len(_inputs), unit="Files", **PBAR_DEFAULTS)

        for src in _inputs:
            dest = output / src.name
            futures[executor.submit(repack_pod5_file, src=src, dest=dest)] = dest

        for future in as_completed(futures):
            tqdm.write(f"Finished {futures[future]}")
            pbar.update(1)

    pbar.close()
    print("Done")


def main():
    run_tool(prepare_pod5_repack_argparser())


if __name__ == "__main__":
    main()



"""
Tool for subsetting pod5 files into one or more outputs using a list of read ids
"""


from pathlib import Path
from time import sleep
from typing import List
from pod5.tools.polars_utils import PL_DEST_FNAME, PL_READ_ID, PL_UUID_REGEX
from pod5.tools.utils import (
    DEFAULT_THREADS,
    PBAR_DEFAULTS,
    collect_inputs,
    init_logging,
    limit_threads,
    logged_all,
)

import polars as pl

from tqdm.auto import tqdm

import pod5 as p5
from pod5.repack import Repacker

from pod5.tools.parsers import prepare_pod5_filter_argparser, run_tool
from pod5.tools.pod5_subset import (
    calculate_transfers,
    parse_sources,
)
from pod5.tools.polars_utils import PL_SRC_FNAME
from pod5.tools.utils import logged


logger = init_logging()

pl.enable_string_cache()


@logged_all
def parse_read_id_targets(ids: Path, output: Path) -> pl.LazyFrame:
    """Parse the list of read_ids checking all are valid uuids"""
    read_ids = (
        pl.scan_csv(
            ids,
            has_header=False,  # Any header will be filtered out by is_uuid
            comment_char="#",
            new_columns=[PL_READ_ID],
            rechunk=False,
        )
        .drop_nulls()
        .unique()
        .with_columns(
            [
                pl.lit(str(output.resolve())).cast(pl.Categorical).alias(PL_DEST_FNAME),
                pl.col(PL_READ_ID).str.contains(PL_UUID_REGEX).alias("is_uuid"),
            ]
        )
        .filter(pl.col("is_uuid"))
        .drop("is_uuid")
    )

    if len(read_ids.fetch(10)) == 0:
        raise AssertionError(f"Found 0 read_ids in {ids}. Nothing to do")

    return read_ids


@logged(log_time=True)
def filter_reads(dest: Path, sources: pl.DataFrame, duplicate_ok: bool) -> None:
    """Copy the reads in `sources` into a new pod5 file at `dest`"""
    repacker = Repacker()
    with p5.Writer(dest) as writer:
        output = repacker.add_output(writer, not duplicate_ok)

        # Count the total number of reads expected
        total_reads = 0
        for source, reads in sources.group_by(PL_SRC_FNAME):
            total_reads += len(reads.get_column(PL_READ_ID))

        pbar = tqdm(
            total=total_reads,
            unit="Read",
            desc="Filtering",
            leave=True,
            **PBAR_DEFAULTS,
        )

        active_limit = 5

        # Copy selected reads from one file at a time
        for source, reads in sources.group_by(PL_SRC_FNAME):
            src = Path(source)
            read_ids = reads.get_column(PL_READ_ID).unique().to_list()
            logger.debug(f"Filtering: {src} - n_reads: {len(read_ids)}")

            if len(read_ids) == 0:
                logger.debug(f"Skipping: {src}")
                continue

            while repacker.currently_open_file_reader_count >= active_limit:
                pbar.update(repacker.reads_completed - pbar.n)
                sleep(0.05)

            with p5.Reader(src) as reader:
                repacker.add_selected_reads_to_output(output, reader, read_ids)
                continue

        repacker.set_output_finished(output)
        while repacker.currently_open_file_reader_count > 0:
            pbar.update(repacker.reads_completed - pbar.n)
            sleep(0.05)

        repacker.finish()
        pbar.close()

    return


@logged_all
def filter_pod5(
    inputs: List[Path],
    output: Path,
    ids: Path,
    missing_ok: bool = False,
    duplicate_ok: bool = False,
    force_overwrite: bool = False,
    recursive: bool = False,
    threads: int = DEFAULT_THREADS,
) -> None:
    """Prepare the pod5 filter mapping and run the repacker"""
    # Remove output file
    if output.exists():
        if not force_overwrite:
            raise FileExistsError(
                f"Output file already exists and --force-overwrite not set - {output}"
            )
        else:
            output.unlink()

    # Create parent directories if they do not exist
    if not output.parent.exists():
        output.parent.mkdir(parents=True, exist_ok=True)

    targets = parse_read_id_targets(ids, output=output)
    print(f"Parsed {len(targets.collect())} reads_ids from: {ids.name}")

    threads = limit_threads(threads)

    _inputs = collect_inputs(inputs, recursive, "*.pod5", threads=threads)
    sources = parse_sources(_inputs, threads=threads)
    print(f"Found {len(sources.collect())} read_ids from {len(_inputs)} inputs")

    # Map the target outputs to which source read ids they're comprised of
    transfers = calculate_transfers(
        sources=sources,
        targets=targets,
        missing_ok=missing_ok,
    )
    print(f"Calculated {len(transfers.collect())} transfers")

    # There will only one output from this
    groupby_dest = transfers.collect().group_by(PL_DEST_FNAME)
    for dest, sources in groupby_dest:
        filter_reads(dest=dest, sources=sources, duplicate_ok=duplicate_ok)

    return


@logged_all
def main():
    """pod5 filter main"""
    run_tool(prepare_pod5_filter_argparser())


if __name__ == "__main__":
    main()



"""
Container class for a pod5 Read object
"""
import datetime
import enum
import os
from dataclasses import dataclass, field
from typing import Dict, List, Union
from uuid import UUID

import numpy as np
import numpy.typing as npt

from pod5.signal_tools import vbz_decompress_signal_chunked

PathOrStr = Union[os.PathLike, str]


class EndReasonEnum(enum.Enum):
    """EndReason Enumeration"""

    UNKNOWN = 0
    MUX_CHANGE = 1
    UNBLOCK_MUX_CHANGE = 2
    DATA_SERVICE_UNBLOCK_MUX_CHANGE = 3
    SIGNAL_POSITIVE = 4
    SIGNAL_NEGATIVE = 5


_END_REASON_FORCED_DEFAULTS: Dict[EndReasonEnum, bool] = {
    EndReasonEnum.UNKNOWN: False,
    EndReasonEnum.MUX_CHANGE: True,
    EndReasonEnum.UNBLOCK_MUX_CHANGE: True,
    EndReasonEnum.DATA_SERVICE_UNBLOCK_MUX_CHANGE: True,
    EndReasonEnum.SIGNAL_POSITIVE: False,
    EndReasonEnum.SIGNAL_NEGATIVE: False,
}


@dataclass(frozen=True)
class EndReason:
    """
    Data on why the Read ended.

    Parameters
    ----------

    reason: EndReasonEnum
        The end reason enumeration.
    forced: bool
        True if it is a 'forced' read break.
    """

    #: The end reason enumeration
    reason: EndReasonEnum
    #: True if it is a 'forced' read break (e.g. mux_change, unblock), False otherwise.
    forced: bool

    @property
    def name(self) -> str:
        """Return the reason name as a lower string"""
        return self.reason.name.lower()

    @classmethod
    def from_reason_with_default_forced(cls, reason: EndReasonEnum) -> "EndReason":
        """
        Return a new EndReason instance with the 'forced' flag set to the expected
        default for the given reason
        """
        return cls(reason=reason, forced=_END_REASON_FORCED_DEFAULTS[reason])


@dataclass()
class Calibration:
    """
    Parameters to convert the signal data to picoamps.

    Parameters
    ----------

    offset: float
        Calibration offset used to convert raw ADC data into pA readings.
    scale: float
        Calibration scale factor used to convert raw ADC data into pA readings.
    """

    #: Calibration offset used to convert raw ADC data into pA readings.
    offset: float
    #: Calibration scale factor used to convert raw ADC data into pA readings.
    scale: float

    @classmethod
    def from_range(
        cls, offset: float, adc_range: float, digitisation: float
    ) -> "Calibration":
        """Create a Calibration instance from offset, adc_range and digitisation"""
        return cls(offset, adc_range / digitisation)


@dataclass()
class Pore:
    """
    Data for the pore that the Read was acquired on

    Parameters
    ----------

    channel: int
        1-indexed channel.
    well: int
        1-indexed well.
    pore_type: PoreType
        The pore type present in the well.
    """

    #: 1-indexed channel.
    channel: int
    #: 1-indexed well.
    well: int
    #: Name of the pore type present in the well.
    pore_type: str


@dataclass(frozen=True)
class RunInfo:
    """
    Higher-level information about the Reads that correspond to a part of an
    experiment, protocol or acquisition

    Parameters
    ----------

    acquisition_id : str
        A unique identifier for the acquisition.
    acquisition_start_time : datetime.datetime
        This is the clock time for sample 0
    adc_max : int
        The maximum ADC value that might be encountered.
    adc_min : int
        The minimum ADC value that might be encountered.
    context_tags : Dict[str, str]
        The context tags for the run. (For compatibility with fast5).
    experiment_name : str
        The user-supplied name for the experiment being run.
    flow_cell_id : str
        Uniquely identifies the flow cell the data was captured on.
    flow_cell_product_code : str
        Identifies the type of flow cell the data was captured on.
    protocol_name : str
        The name of the protocol that was run.
    protocol_run_id : str
        The unique identifier for the protocol run that produced this data.
    protocol_start_time : datetime.datetime
         When the protocol that the acquisition was part of started.
    sample_id : str
        A user-supplied name for the sample being analysed.
    sample_rate : int
        The number of samples acquired each second on each channel.
    sequencing_kit : str
        The type of sequencing kit used to prepare the sample.
    sequencer_position : str
        The sequencer position the data was collected on.
    sequencer_position_type : str
        The type of sequencing hardware the data was collected on.
    software : str
        A description of the software that acquired the data.
    system_name : str
        The name of the system the data was collected on.
    system_type : str
        The type of system the data was collected on.
    tracking_id : Dict[str, str]
        The tracking id for the run. (For compatibility with fast5).

    """

    #: A unique identifier for the acquisition - note that readers should not
    #: depend on this uniquely determining the other fields in the run_info, or being
    #: unique among the dictionary keys.
    acquisition_id: str
    #: This is the clock time for sample 0
    acquisition_start_time: datetime.datetime
    #: The maximum ADC value that might be encountered. This is a hardware constraint.
    adc_max: int
    #: The minimum ADC value that might be encountered. This is a hardware constraint.
    adc_min: int
    #: The context tags for the run. (For compatibility with fast5).
    context_tags: Dict[str, str] = field(hash=False, compare=True)
    #: The user-supplied name for the experiment being run.
    experiment_name: str
    #: Uniquely identifies the flow cell the data was captured on.
    #: This is written on the flow cell case.
    flow_cell_id: str
    #: Identifies the type of flow cell the data was captured on.
    flow_cell_product_code: str
    #: The name of the protocol that was run.
    protocol_name: str
    #: The unique identifier for the protocol run that produced this data.
    protocol_run_id: str
    #:  When the protocol that the acquisition was part of started.
    protocol_start_time: datetime.datetime
    #: A user-supplied name for the sample being analysed.
    sample_id: str
    #: The number of samples acquired each second on each channel.
    sample_rate: int
    #: The type of sequencing kit used to prepare the sample.
    sequencing_kit: str
    #: The sequencer position the data was collected on. For removable positions,
    #: like MinION Mk1Bs, this is unique (e.g. 'MN12345'), while for integrated
    #: positions it is not (e.g. 'X1' on a GridION).
    sequencer_position: str
    #: The type of sequencing hardware the data was collected on. For example:
    #: 'MinION Mk1B' or 'GridION' or 'PromethION'.
    sequencer_position_type: str
    #: A description of the software that acquired the data. For example:
    #: 'MinKNOW 21.05.12 (Bream 5.1.6, Configurations 16.2.1, Core 5.1.9, Guppy 4.2.3)'.
    software: str
    #: The name of the system the data was collected on. This might be a sequencer
    #: serial (eg: 'GXB1234') or a host name (e.g. 'Lab PC').
    system_name: str
    #: The type of system the data was collected on. For example, 'GridION Mk1' or
    #: 'PromethION P48'. If the system is not a Nanopore sequencer with built-in
    #: compute, this will be a description of the operating system
    #: (e.g. 'Ubuntu 20.04').
    system_type: str
    #: The tracking id for the run. (For compatibility with fast5).
    tracking_id: Dict[str, str] = field(hash=False, compare=True)


@dataclass()
class ShiftScalePair:
    """A pair of floating point shift and scale values."""

    shift: float = field(default=float("nan"))
    scale: float = field(default=float("nan"))


@dataclass()
class BaseRead:
    """
    Base class for POD5 Read Data

    Parameters
    ----------

    read_id : UUID
        The read_id of this read as UUID.
    pore : Pore
        Pore data.
    calibration : Calibration
        Calibration data.
    read_number : int
        The read number on channel. This is increasing but typically
        not necessarily consecutive.
    start_sample : int
        The number samples recorded on this channel before the read started.
    median_before : float
        The level of current in the well before this read.
    end_reason : EndReason
        EndReason data.
    run_info : RunInfo
        RunInfo data.
    num_minknow_events: int
        Number of minknow events that the read contains
    tracked_scaling: ShiftScalePair
        Shift and Scale for tracked read scaling values (based on previous reads shift)
    predicted_scaling: ShiftScalePair
        Shift and Scale for predicted read scaling values (based on this read's raw signal)
    num_reads_since_mux_change: int
        Number of selected reads since the last mux change on this reads channel
    time_since_mux_change: float
        Time in seconds since the last mux change on this reads channel
    """

    #: The read_id of this read as UUID
    read_id: UUID
    #: Pore metadata
    pore: Pore
    #: Calibration metadata
    calibration: Calibration
    #: The read number on channel. This is increasing but typically
    #: not necessarily consecutive.
    read_number: int
    #: The number samples recorded on this channel before the read started.
    start_sample: int
    #: The level of current in the well before this read.
    median_before: float
    #: EndReason data.
    end_reason: EndReason
    #: RunInfo data.
    run_info: RunInfo
    #: Number of minknow events that the read contains
    num_minknow_events: int = field(default=0)
    #: Shift and Scale for tracked read scaling values (based on previous reads shift)
    tracked_scaling: ShiftScalePair = field(default_factory=ShiftScalePair)
    #: Shift and Scale for predicted read scaling values (based on this read's raw signal)
    predicted_scaling: ShiftScalePair = field(default_factory=ShiftScalePair)
    #: Number of selected reads since the last mux change on this reads channel
    num_reads_since_mux_change: int = field(default=0)
    #: Time in seconds since the last mux change on this reads channel
    time_since_mux_change: float = field(default=0.0)


@dataclass()
class Read(BaseRead):
    """
    POD5 Read Data with an uncompressed signal

    Parameters
    ----------

    read_id : UUID
        The read_id of this read as UUID.
    pore : Pore
        Pore data.
    calibration : Calibration
        Calibration data.
    read_number : int
        The read number on channel. This is increasing but typically
        not necessarily consecutive.
    start_sample : int
        The number samples recorded on this channel before the read started.
    median_before : float
        The level of current in the well before this read.
    end_reason : EndReason
        EndReason data.
    run_info : RunInfo
        RunInfo data.
    signal : numpy.array[int16]
        Uncompressed signal data.
    """

    #: Uncompressed signal data.
    signal: npt.NDArray[np.int16] = field(
        default_factory=lambda: np.array([], dtype=np.int16)
    )

    @property
    def sample_count(self) -> int:
        """Return the total number of samples in the uncompressed signal."""
        return len(self.signal)


@dataclass()
class CompressedRead(BaseRead):
    """
    POD5 Read Data with a compressed signal.

    Parameters
    ----------

    read_id : UUID
        The read_id of this read as UUID.
    pore : Pore
        Pore data.
    calibration : Calibration
        Calibration data.
    read_number : int
        The read number on channel. This is increasing but typically
        not necessarily consecutive.
    start_sample : int
        The number samples recorded on this channel before the read started.
    median_before : float
        The level of current in the well before this read.
    end_reason : EndReason
        EndReason data.
    run_info : RunInfo
        RunInfo data.
    signal_chunks : List[numpy.array[uint8]]
        Compressed signal data in chunks.
    signal_chunk_lengths : List[int]
        Chunk lengths (number of samples) of signal data **before** compression.
    """

    #: Compressed signal data in chunks.
    signal_chunks: List[npt.NDArray[np.uint8]] = field(default_factory=list)

    #: Chunk lengths (number of samples) of signal data **before** compression.
    signal_chunk_lengths: List[int] = field(default_factory=list)

    @property
    def sample_count(self) -> int:
        """Return the total number of samples in the uncompressed signal."""
        return sum(self.signal_chunk_lengths)

    @property
    def decompressed_signal(self) -> npt.NDArray[np.int16]:
        """
        Decompress and return the chunked signal data as a contiguous numpy array.

        Returns
        -------
        decompressed_signal : numpy.array[int16]
            Decompressed signal data
        """
        return vbz_decompress_signal_chunked(
            self.signal_chunks, self.signal_chunk_lengths
        )



"""
Tool for updating pod5 files to the latest available version
"""
from typing import Iterable
from pathlib import Path

from tqdm.auto import tqdm

import lib_pod5 as p5b

import pod5 as p5
from pod5.tools.parsers import prepare_pod5_update_argparser, run_tool
from pod5.tools.utils import (
    PBAR_DEFAULTS,
    assert_no_duplicate_filenames,
    collect_inputs,
)


def update_pod5(
    inputs: Iterable[Path],
    output: Path,
    force_overwrite: bool = False,
    recursive: bool = False,
):
    """
    Given a list of pod5 files, update their tables to the most recent version
    """
    if not output.exists():
        output.mkdir(parents=True, exist_ok=True)

    paths = collect_inputs(inputs, recursive=recursive, pattern="*.pod5")
    assert_no_duplicate_filenames(paths)

    exists = set(output / p.name for p in paths if Path(output / p.name).exists())

    if not paths.isdisjoint(exists):
        inout = [p.name for p in exists - paths]
        raise AssertionError(f"Cannot update inputs in-place. Found: {inout}")

    if not force_overwrite and exists:
        raise FileExistsError(
            f"{len(exists)} Output files already exists and --force-overwrite not set. "
            f"Found: {exists}"
        )
    else:
        for path in exists:
            path.unlink()

    pbar = tqdm(
        total=len(paths), desc="Updating", unit="File", leave=True, **PBAR_DEFAULTS
    )

    for path in paths:
        dest = output / path.name
        with p5.Reader(path) as reader:
            p5b.update_file(reader.inner_file_reader, str(dest))
        pbar.update()


def main():
    run_tool(prepare_pod5_update_argparser())


if __name__ == "__main__":
    main()



"""POD5 Format

Bindings for the POD5 file format
"""


# Pull the version from the pyproject.toml
import sys

if sys.version_info >= (3, 8):
    from importlib import metadata
else:
    import importlib_metadata as metadata

__version__ = metadata.version("pod5")

from .api_utils import (
    format_read_id_to_str,
    format_read_ids,
    load_read_id_iterable,
    pack_read_ids,
)
from .pod5_types import (
    Calibration,
    CompressedRead,
    EndReason,
    EndReasonEnum,
    Pore,
    Read,
    RunInfo,
)
from .reader import Reader, ReadRecord, ReadRecordBatch
from .dataset import DatasetReader
from .signal_tools import (
    vbz_compress_signal,
    vbz_decompress_signal,
    vbz_decompress_signal_chunked,
    vbz_decompress_signal_into,
)
from .writer import Writer

__all__ = (
    "__version__",
    "format_read_id_to_str",
    "format_read_ids",
    "load_read_id_iterable",
    "pack_read_ids",
    "DatasetReader",
    "Calibration",
    "CompressedRead",
    "EndReason",
    "EndReasonEnum",
    "Pore",
    "Read",
    "RunInfo",
    "Reader",
    "ReadRecord",
    "ReadRecordBatch",
    "vbz_compress_signal",
    "vbz_decompress_signal",
    "vbz_decompress_signal_chunked",
    "vbz_decompress_signal_into",
    "Writer",
)



"""
Tools for handling pod5 signals
"""

from typing import List, Tuple, Union

import lib_pod5 as p5b
import numpy as np
import numpy.typing as npt

DEFAULT_SIGNAL_CHUNK_SIZE = 102400


def vbz_decompress_signal(
    compressed_signal: Union[npt.NDArray[np.uint8], memoryview], sample_count: int
) -> npt.NDArray[np.int16]:
    """
    Decompress a contiguous (not-chunked) numpy array of compressed signal data

    Parameters
    ----------
    compressed_signal : numpy.ndarray[uint8]
        The array of compressed signal data to decompress.
    sample_count : int
        The number of samples in the original signal

    Returns
    -------
    A decompressed signal array numpy.ndarray[int16]
    """
    if len(compressed_signal) == 0:
        return np.array([], dtype=np.int16)

    signal = np.empty(sample_count, dtype="i2")
    p5b.decompress_signal(compressed_signal, signal)
    return signal


def vbz_decompress_signal_chunked(
    compressed_signal_chunks: List[npt.NDArray[np.uint8]], sample_counts: List[int]
) -> npt.NDArray[np.int16]:
    """
    Decompress a chunks of numpy array of compressed signal data

    Parameters
    ----------
    compressed_signal_chunks : List[numpy.ndarray[uint8]]
        A list of compressed signal data chunks to decompress.
    sample_counts : List[int]
        The number of samples in the original signal chunks

    Returns
    -------
    A decompressed signal array numpy.ndarray[int16]

    Raises
    ------
    ValueError
        Inconsistent parameter lengths
    """
    if len(compressed_signal_chunks) != len(sample_counts):
        raise ValueError(
            f"Inconsistent number of chunks to decompress - "
            f"signals: {len(compressed_signal_chunks)}, counts: {len(sample_counts)}"
        )

    if len(compressed_signal_chunks) == 0:
        return np.array([], dtype=np.int16)

    decompressed_signal: npt.NDArray[
        np.int16
    ] = np.concatenate(  # type:ignore [no-untyped-call]
        [
            vbz_decompress_signal(signal_chunk, sample_count)
            for signal_chunk, sample_count in zip(
                compressed_signal_chunks, sample_counts
            )
        ]
    )
    return decompressed_signal


def vbz_decompress_signal_into(
    compressed_signal: Union[npt.NDArray[np.uint8], memoryview],
    output_array: npt.NDArray[np.int16],
) -> npt.NDArray[np.int16]:
    """
    Decompress a numpy array of compressed signal data into the destination
    "output_array"

    Parameters
    ----------
    compressed_signal : numpy.ndarray[uint8]
        The array of compressed signal data to decompress.
    output_array : numpy.ndarray[int16]
        The destination location for signal

    Returns
    -------
    A decompressed signal array numpy.ndarray[int16]
    """
    if len(compressed_signal) == 0:
        return np.array([], dtype=np.int16)

    p5b.decompress_signal(compressed_signal, output_array)
    return output_array


def vbz_compress_signal(signal: npt.NDArray[np.int16]) -> npt.NDArray[np.uint8]:
    """
    Compress a numpy array of signal data

    Parameters
    ----------
    signal : numpy.ndarray[int16]
        The array of signal data to compress.

    Returns
    -------
    compressed_signal : numpy.array[uint8]
        The compressed signal data as a numpy.ndarray[uint8] (byte array)
    """
    if signal.size == 0:
        return np.array([], dtype=np.uint8)

    max_signal_size = p5b.vbz_compressed_signal_max_size(len(signal))
    compressed_signal = np.zeros(max_signal_size, dtype="u1")

    size = p5b.compress_signal(signal, compressed_signal)

    return np.resize(compressed_signal, size)


def vbz_compress_signal_chunked(
    signal: npt.NDArray[np.int16], signal_chunk_size: int = DEFAULT_SIGNAL_CHUNK_SIZE
) -> Tuple[List[npt.NDArray[np.uint8]], List[int]]:
    """
    Compress a numpy array of signal data into chunks

    Parameters
    ----------
    signal : numpy.ndarray[int16]
        The array of signal data to compress.
    signal_chunk_size : int
        The number of signal samples in a chunk

    Returns
    -------
    compressed_signal_chunks : List[numpy.array[uint8]]
        A List of chunks of compressed signal data as numpy.ndarray[uint8] (byte arrays)
    signal_chunk_lengths : List[int]
        The number of uncompressed signal samples in each chunk
    """
    signal_chunks: List[npt.NDArray[np.uint8]] = []
    signal_chunk_lengths: List[int] = []

    # Take slice views of the signal ndarray (non-copying)
    for slice_index in range(0, len(signal), signal_chunk_size):
        signal_slice = signal[slice_index : slice_index + signal_chunk_size]
        signal_chunks.append(vbz_compress_signal(signal_slice))
        signal_chunk_lengths.append(len(signal_slice))

    return signal_chunks, signal_chunk_lengths



"""
Utility functions for the pod5 API
"""


import warnings
from typing import Any, Collection, List, Union

import numpy as np
import numpy.typing as npt
import pyarrow as pa
from lib_pod5 import format_read_id_to_str, load_read_id_iterable


class Pod5ApiException(Exception):
    """Generic Pod5 API Exception"""


def pack_read_ids(
    read_ids: Collection[str], invalid_ok: bool = False
) -> npt.NDArray[np.uint8]:
    """
    Convert a `Collection` of `read_id` strings to a `numpy.ndarray`
    in preparation for writing to pod5 files.

    Parameters
    ----------
    read_ids : Collection[str]
        Collection of well-formatted read_id strings

    Returns
    -------
    packed_read_ids : numpy.ndarray[uint8]
        Repacked read_ids ready for writing to pod5 files.
    """
    read_id_data = np.empty(shape=(len(read_ids), 16), dtype=np.uint8)
    count = load_read_id_iterable(read_ids, read_id_data)
    if invalid_ok is False and count != len(read_ids):
        raise RuntimeError("Invalid read id passed")

    return read_id_data


def format_read_ids(
    read_ids: Union[npt.NDArray[np.uint8], pa.lib.FixedSizeBinaryArray]
) -> List[str]:
    """
    Convert a packed array of read_ids and convert them to a list of strings.

    Parameters
    ----------
    read_ids : numpy.ndarray[uint8], pa.lib.FixedSizeBinaryArray
        Packed read_ids from a numpy.ndarray or read directly from pod5 file

    Returns
    -------
    read_ids : list[str]
        A list of converted read_ids as strings
    """
    if isinstance(read_ids, pa.lib.FixedSizeBinaryArray):
        read_ids = read_ids.buffers()[1]
    return format_read_id_to_str(read_ids)


def deprecation_warning(deprecated: str, alternate: str) -> None:
    """
    Issue a `FutureWarning` warning that `deprecated` has been deprecated in favour of
    `alternate`.

    Parameters
    ----------
    deprecated : str
        The module path to the deprecated item
    alternate : str
        The module path to the alternate item
    """
    warnings.warn(
        f"{deprecated} is deprecated. Please use {alternate}",
        DeprecationWarning,
        stacklevel=2,
    )


def safe_close(obj: Any, attr: str) -> None:
    """
    Try to close() an object's attribute ignoring any exceptions raised.
    This is used to safely handle closing potentially unassigned attributes
    while calling close() in __del__()
    """
    if not hasattr(obj, attr):
        return

    try:
        getattr(obj, attr).close()
    except Exception:
        pass



"""
Tool for subsetting pod5 files into one or more outputs
"""

from copy import deepcopy
import multiprocessing as mp
from multiprocessing.context import SpawnContext
from pathlib import Path
from queue import Empty
from string import Formatter
import sys
from time import sleep
from typing import Any, List, Optional, Set, Tuple

import polars as pl
from tqdm.auto import tqdm
import pod5 as p5
import pod5.repack as p5_repack
from pod5.tools.polars_utils import (
    PL_DEST_FNAME,
    PL_READ_ID,
    PL_SRC_FNAME,
    PL_UUID_REGEX,
    pl_format_read_id,
)
from pod5.tools.utils import (
    DEFAULT_THREADS,
    PBAR_DEFAULTS,
    collect_inputs,
    init_logging,
    limit_threads,
    logged,
    logged_all,
    terminate_processes,
)
from pod5.tools.parsers import prepare_pod5_subset_argparser, run_tool


DEFAULT_READ_ID_COLUMN = "read_id"

logger = init_logging()

pl.enable_string_cache()


@logged_all
def get_separator(path: Path) -> str:
    """
    Inspect the first line of the file at path and attempt to determine the field
    separator as either tab or comma, depending on the number of occurrences of each
    Returns "," or "<tab>"
    """
    with path.open("r") as fh:
        line = fh.readline()
    n_tabs = line.count("\t")
    n_comma = line.count(",")
    if n_tabs >= n_comma:
        return "\t"
    return ","


@logged_all
def default_filename_template(subset_columns: List[str]) -> str:
    """Create the default filename template from the subset_columns selected"""
    default = "_".join(f"{col}-{{{col}}}" for col in subset_columns)
    default += ".pod5"
    return default


@logged_all
def column_keys_from_template(template: str) -> List[str]:
    """Get a list of placeholder keys in the template"""
    return [key for _, key, _, _ in Formatter().parse(template) if key]


@logged_all
def fstring_to_polars(
    template: str,
) -> Tuple[str, List[str]]:
    """
    Repalce f-string keyed placeholders with positional ones and return the keys in
    their respective position
    """
    # This is for pl.format positional syntax
    replaced = template
    keys = column_keys_from_template(template)
    for key in keys:
        replaced = replaced.replace(f"{{{key}}}", "{}")
    return replaced, keys


@logged_all
def parse_table_mapping(
    summary_path: Path,
    filename_template: Optional[str],
    subset_columns: List[str],
    read_id_column: str = DEFAULT_READ_ID_COLUMN,
    ignore_incomplete_template: bool = False,
) -> pl.LazyFrame:
    """
    Parse a table using polars to create a mapping of output targets to read ids
    """
    if not subset_columns:
        raise AssertionError("Missing --columns when using --summary / --table")

    if not filename_template:
        filename_template = default_filename_template(subset_columns)

    assert_filename_template(
        filename_template, subset_columns, ignore_incomplete_template
    )

    # Add the destination filename as a column
    pl_template, keys = fstring_to_polars(filename_template)

    columns = deepcopy(subset_columns)
    columns.append(read_id_column)

    targets = (
        pl.read_csv(
            summary_path,
            columns=columns,
            separator=get_separator(summary_path),
            comment_char="#",
        )
        .lazy()
        .with_columns(
            [
                pl.format(pl_template, *keys).cast(pl.Categorical).alias(PL_DEST_FNAME),
                pl.col(read_id_column).alias(PL_READ_ID),
            ]
        )
    )
    return targets


@logged_all
def assert_filename_template(
    template: str, subset_columns: List[str], ignore_incomplete_template: bool
) -> None:
    """
    Get the keys named in the template to assert that they exist in subset_columns
    """
    # Parse the template string to get the keywords
    # "{hello}_world_{name}" -> ["hello", "name"]
    template_keys = set(args[1] for args in Formatter().parse(template) if args[1])
    allowed_keys = set(subset_columns)

    # Assert there are no unexpected keys in the template
    unexpected = template_keys - allowed_keys
    if unexpected:
        raise KeyError(f"--template {template} has unexpected keys: {unexpected}")

    # Assert there are no unused keys in the template
    # This is important as the output would be degenerate on some keys
    if not ignore_incomplete_template:
        unused = allowed_keys - template_keys
        if unused:
            raise KeyError(
                f"--template {template} does not use {unused} keys. "
                "Use --ignore-incomplete-template to suppress this exception."
            )


@logged_all
def create_default_filename_template(subset_columns: List[str]) -> str:
    """Create the default filename template from the subset_columns selected"""
    default = "_".join(f"{col}-{{{col}}}" for col in subset_columns)
    default += ".pod5"
    return default


@logged_all
def parse_csv_mapping(csv_path: Path) -> pl.LazyFrame:
    """Parse the csv direct mapping of output target to read_ids to a targets dataframe"""
    targets = (
        pl.scan_csv(
            csv_path,
            has_header=False,
            comment_char="#",
            new_columns=[PL_DEST_FNAME, PL_READ_ID],
            rechunk=False,
        )
        .drop_nulls()
        .with_columns(
            [
                pl.col(PL_DEST_FNAME).cast(pl.Categorical),
                pl.col(PL_READ_ID).str.contains(PL_UUID_REGEX).alias("is_uuid"),
            ]
        )
        .filter(pl.col("is_uuid"))
        .drop("is_uuid")
    )

    if len(targets.fetch(10)) == 0:
        raise AssertionError(f"Found 0 read_ids in {csv_path}. Nothing to do")

    return targets


@logged_all
def resolve_output_targets(targets: pl.LazyFrame, output: Path) -> pl.LazyFrame:
    """Prepend the output path to the target filename and resolve the complete string"""
    # Add output directory column
    # Concatenate output directory to the output filenmae, drop temporary column
    targets = targets.with_columns(
        pl.concat_str(
            [
                pl.lit(str(output.resolve())).cast(pl.Categorical),
                pl.col(PL_DEST_FNAME).cast(pl.Categorical),
            ],
            separator="/",
        )
        .cast(pl.Categorical)
        .alias(PL_DEST_FNAME),
    )

    return targets


@logged_all
def assert_overwrite_ok(targets: pl.LazyFrame, force_overwrite: bool) -> None:
    """
    Given the target filenames, assert that no unforced overwrite will occur
    unless requested raising an FileExistsError. Unlinks existing files if they exist
    if `force_overwrite` set
    """

    def exists(path: str) -> bool:
        return Path(path).exists()

    DEST_EXISTS = "__dest_exists"
    dests = (
        targets.select(pl.col(PL_DEST_FNAME).unique())
        .with_columns(pl.col(PL_DEST_FNAME).apply(exists).alias(DEST_EXISTS))
        .collect()
    )

    if dests.get_column(DEST_EXISTS).any():
        if not force_overwrite:
            raise FileExistsError(
                "Output files already exists and --force-overwrite not set. "
            )

        def unlinker(item) -> bool:
            path, *_ = item
            Path(path).unlink()
            return True

        dests.select(pl.col(PL_DEST_FNAME).where(pl.col(DEST_EXISTS))).apply(unlinker)


@logged_all
def parse_source_process(paths: mp.JoinableQueue, parsed_sources: mp.Queue):
    """Parse sources until paths queue is consumed"""
    while True:
        path = paths.get(timeout=60)

        # Seninel value of finished work
        if path is None:
            paths.task_done()
            break

        parsed_sources.put(parse_source(path))
        paths.task_done()

    parsed_sources.close()
    paths.close()


@logged_all
def parse_source(path: Path) -> pl.LazyFrame:
    """
    Reads the read ids available in a given pod5 file returning a dataframe
    with the formatted read_ids and the source filename
    """
    with p5.Reader(path) as rdr:
        pa_read_table = rdr.read_table.read_all()

    source = (
        pl.from_arrow(pa_read_table, rechunk=False)
        .with_columns(
            pl_format_read_id(pl.col("read_id")).alias(PL_READ_ID),
            pl.lit(str(path.resolve())).alias(PL_SRC_FNAME),
        )
        .select(PL_READ_ID, PL_SRC_FNAME)
        .lazy()
    )
    return source


@logged_all
def parse_sources(paths: Set[Path], threads: int = DEFAULT_THREADS) -> pl.LazyFrame:
    """Reads all inputs and return formatted lazy dataframe"""

    threads = limit_threads(threads)
    n_proc = min(threads, len(paths))

    ctx = mp.get_context("spawn")
    work: mp.JoinableQueue = ctx.JoinableQueue(maxsize=len(paths) + n_proc)
    parsed_sources = ctx.Queue(maxsize=len(paths))
    for path in paths:
        work.put(path)

    # Spawn worker processes
    active_processes = []
    for _ in range(n_proc):
        process = ctx.Process(
            target=parse_source_process,
            args=(work, parsed_sources),
            daemon=True,
        )
        # Enqueue a sentinel for each process to stop
        work.put(None)

        process.start()
        active_processes.append(process)

    # Wait for all work to be done
    work.join()
    work.close()

    items: List[pl.LazyFrame] = []
    for _ in range(len(paths)):
        # After work is joined we will have len(paths) items in parsed_sources
        # shouldn't need to wait or seconds or ever see Empty.
        items.append(parsed_sources.get(timeout=60))

    parsed_sources.close()
    parsed_sources.join_thread()

    sources = pl.concat(
        items=items,
        how="vertical",
        rechunk=False,
        parallel=True,
    )

    # Shutdown
    for proc in active_processes:
        proc.join()
        proc.close()

    return sources


@logged_all
def calculate_transfers(
    sources: pl.LazyFrame, targets: pl.LazyFrame, missing_ok: bool
) -> pl.LazyFrame:
    """
    Produce the transfers dataframe which maps the read_ids, source and destination
    """
    transfers = targets.join(sources, on=PL_READ_ID, how="left").select(
        PL_READ_ID, PL_SRC_FNAME, PL_DEST_FNAME
    )

    if not missing_ok:
        # Find any records where there there is no source i.e. input is missing
        if (
            transfers.select(pl.col(PL_SRC_FNAME).is_null().any())
            .collect()
            .to_series()
            .any()
        ):
            raise AssertionError(
                "Missing read_ids from inputs but --missing-ok not set"
            )

    # Add filter to transfers query removing missing inputs
    transfers = transfers.filter(pl.col(PL_SRC_FNAME).is_not_null())
    return transfers


class WorkQueue:
    def __init__(
        self,
        context: SpawnContext,
        transfers: pl.LazyFrame,
    ) -> None:
        self.work: mp.JoinableQueue = context.JoinableQueue()
        self.size = 0
        groupby_dest = transfers.collect().group_by(PL_DEST_FNAME)
        for dest, sources in groupby_dest:
            self.work.put((Path(dest), sources))
            self.size += 1

        self.progress: mp.Queue = context.Queue(maxsize=self.size + 1)
        logger.info(f"WorkQueue size: {self.size}")

    @logged_all
    def join(self) -> None:
        """Call join on the work queue waiting for all tasks to be done"""
        self.work.join()

    @logged_all
    def _discard_and_close(self, queue: mp.Queue) -> int:
        """
        Discard all remaining enqueued items and close a queue to nicely shutdown the
        queue. Returns the number of discarded items
        """
        count = 0
        while True:
            try:
                queue.get(timeout=0.1)
                count += 1
            except Exception:
                break
        queue.close()
        queue.join_thread()
        return count

    @logged_all
    def shutdown(self) -> int:
        """Shutdown all queues returning the counts of all remaining items"""
        n_work = self._discard_and_close(self.work)

        if n_work > 0:
            print("Unfinished work remaining during shutdown!", file=sys.stderr)

        return n_work


@logged_all
def overall_progress(queue: WorkQueue):
    pbar = tqdm(
        total=queue.size,
        desc="Subsetting",
        unit="Files",
        leave=True,
        position=0,
        **PBAR_DEFAULTS,
    )

    count = 0
    while count < queue.size:
        try:
            queue.progress.get(timeout=1)
            count += 1
            pbar.update()
        except Empty:
            continue


@logged_all
def launch_subsetting(
    transfers: pl.LazyFrame, duplicate_ok: bool, threads: int = DEFAULT_THREADS
) -> None:
    """
    Iterate over the transfers dataframe subsetting reads from sources to destinations
    """
    threads = limit_threads(threads)
    assert {PL_READ_ID, PL_SRC_FNAME, PL_DEST_FNAME}.issubset(set(transfers.columns))

    ctx = mp.get_context("spawn")
    work = WorkQueue(ctx, transfers)

    active_processes = []
    try:
        # Spawn worker processes
        for idx in range(min(threads, work.size)):
            process = ctx.Process(
                target=process_subset_tasks,
                args=(work, idx + 1, duplicate_ok),
                daemon=True,
            )
            # Enqueue a sentinel for each process to stop
            work.work.put(None)
            process.start()
            active_processes.append(process)

        # Spawn progressbar process
        progress_proc = ctx.Process(
            target=overall_progress,
            args=(work,),
            daemon=True,
        )
        progress_proc.start()
        active_processes.append(progress_proc)

        # Wait for all work to be done
        work.join()
        work.shutdown()

        # Shutdown
        for proc in active_processes:
            proc.join()
            proc.close()

    except Exception as exc:
        terminate_processes(active_processes)
        raise exc


@logged(log_time=True)
def process_subset_tasks(queue: WorkQueue, process: int, duplicate_ok: bool):
    """Consumes work from the queue and launches subsetting tasks"""
    while True:
        task = queue.work.get(timeout=60)
        if task is None:
            queue.work.task_done()
            break

        target, sources = task
        try:
            subset_reads(target, sources, process, duplicate_ok)
        finally:
            queue.work.task_done()
            queue.progress.put(True)


@logged(log_time=True)
def subset_reads(
    dest: Path, sources: pl.DataFrame, process: int, duplicate_ok: bool
) -> None:
    """Copy the reads in `sources` into a new pod5 file at `dest`"""
    # Count the total number of reads expected
    total_reads = 0
    for source, reads in sources.group_by(PL_SRC_FNAME):
        total_reads += len(reads.get_column(PL_READ_ID))

    pbar = tqdm(
        total=total_reads,
        desc=dest.name,
        unit="Reads",
        leave=False,
        position=process,
        delay=2,
        **PBAR_DEFAULTS,
    )

    repacker = p5_repack.Repacker()
    with p5.Writer(dest) as writer:
        output = repacker.add_output(writer, not duplicate_ok)

        active_limit = 5
        # Copy selected reads from one file at a time
        for source, reads in sources.group_by(PL_SRC_FNAME):
            while repacker.currently_open_file_reader_count >= active_limit:
                pbar.update(repacker.reads_completed - pbar.n)
                sleep(0.05)

            read_ids = reads.get_column(PL_READ_ID).unique().to_list()
            logger.debug(f"Subsetting: {source} - n_reads: {len(read_ids)}")

            with p5.Reader(Path(source)) as reader:
                repacker.add_selected_reads_to_output(output, reader, read_ids)

        repacker.set_output_finished(output)
        while repacker.currently_open_file_reader_count > 0:
            pbar.update(repacker.reads_completed - pbar.n)
            sleep(0.1)

        pbar.update(total_reads - pbar.n)

        # Finish the pod5 file and close source handles
        repacker.finish()

    pbar.close()

    return


@logged(log_time=True)
def subset_pod5s_with_mapping(
    inputs: Set[Path],
    output: Path,
    targets: pl.LazyFrame,
    threads: int = DEFAULT_THREADS,
    missing_ok: bool = False,
    duplicate_ok: bool = False,
    force_overwrite: bool = False,
) -> None:
    """
    Given an iterable of input pod5 paths and an output directory, create output pod5
    files containing the read_ids specified in the given mapping of output filename to
    set of read_id.
    """

    if not output.exists():
        output.mkdir(parents=True, exist_ok=True)

    targets = resolve_output_targets(targets, output)
    assert_overwrite_ok(targets, force_overwrite)

    print(f"Parsed {len(targets.collect())} targets")
    sources_df = parse_sources(inputs, threads)

    transfers = calculate_transfers(
        sources=sources_df,
        targets=targets,
        missing_ok=missing_ok,
    )

    print(f"Calculated {len(transfers.collect())} transfers")
    launch_subsetting(transfers=transfers, duplicate_ok=duplicate_ok, threads=threads)

    print("Done")
    return None


@logged(log_time=True)
def subset_pod5(
    inputs: List[Path],
    output: Path,
    columns: List[str],
    csv: Optional[Path] = None,
    table: Optional[Path] = None,
    threads: int = DEFAULT_THREADS,
    template: str = "",
    read_id_column: str = DEFAULT_READ_ID_COLUMN,
    missing_ok: bool = False,
    duplicate_ok: bool = False,
    ignore_incomplete_template: bool = False,
    force_overwrite: bool = False,
    recursive: bool = False,
) -> Any:
    """Prepare the subsampling mapping and run the repacker"""

    if csv:
        targets = parse_csv_mapping(csv)

    elif table:
        targets = parse_table_mapping(
            table, template, columns, read_id_column, ignore_incomplete_template
        )

    else:
        raise RuntimeError(
            "Arguments provided could not be used to generate a subset mapping."
        )

    if not output.exists():
        output.mkdir(parents=True)

    _inputs = collect_inputs(
        inputs, recursive=recursive, pattern="*.pod5", threads=threads
    )
    if len(_inputs) == 0:
        raise ValueError("Found no input pod5 files")

    subset_pod5s_with_mapping(
        inputs=_inputs,
        output=output,
        targets=targets,
        threads=threads,
        missing_ok=missing_ok,
        duplicate_ok=duplicate_ok,
        force_overwrite=force_overwrite,
    )


@logged()
def main():
    """pod5 subsample main"""
    run_tool(prepare_pod5_subset_argparser())


if __name__ == "__main__":
    main()



"""
Tools for writing POD5 data
"""
import datetime
import itertools
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import lib_pod5 as p5b
import numpy as np
import pytz

from pod5.api_utils import Pod5ApiException, safe_close
from pod5.pod5_types import (
    BaseRead,
    CompressedRead,
    EndReason,
    PathOrStr,
    Read,
    RunInfo,
)

DEFAULT_SOFTWARE_NAME = "Python API"

PoreType = str
T = TypeVar("T", bound=Union[EndReason, PoreType, RunInfo])


def force_type_and_default(value, dtype, count, default_value=None):
    if default_value is not None and value is None:
        value = np.array([default_value] * count, dtype=dtype)
    assert value is not None
    return value.astype(type, copy=False)


def map_to_tuples(info_map: Any) -> List[Tuple[str, str]]:
    """
    Convert a fast5 property map (e.g. context_tags and tracking_id) to a
    tuple or string pairs to pass to pod5 C API
    """
    if isinstance(info_map, dict):
        return list((str(key), str(value)) for key, value in info_map.items())
    if isinstance(info_map, list):
        return list((str(item[0]), str(item[1])) for item in info_map)
    raise TypeError(f"Unknown input type for context tags {type(info_map)}")


def timestamp_to_int(time_stamp: Union[datetime.datetime, int]) -> int:
    """Convert a datetime timestamp to an integer if it's not already an integer"""
    if isinstance(time_stamp, int):
        return time_stamp
    return int(time_stamp.astimezone(pytz.utc).timestamp() * 1000)


class Writer:
    """Pod5 File Writer"""

    def __init__(self, path: PathOrStr, software_name: str = DEFAULT_SOFTWARE_NAME):
        """
        Open a pod5 file for Writing.

        Parameters
        ----------
        path : os.PathLike, str
            The path to the pod5 file to create
        software_name : str
            The name of the application used to create this pod5 file
        """
        self._path = Path(path).absolute()
        self._software_name = software_name

        if self._path.is_file():
            raise FileExistsError(
                f"Input path already exists. Refusing to overwrite: {self._path}"
            )

        self._writer: Optional[p5b.FileWriter] = p5b.create_file(
            str(self._path), software_name, None
        )
        if not self._writer:
            raise Pod5ApiException(
                f"Failed to open writer at {self._path} : {p5b.get_error_string()}"
            )

        self._end_reasons: Dict[EndReason, int] = {}
        self._pores: Dict[PoreType, int] = {}
        self._run_infos: Dict[RunInfo, int] = {}

        # Internal lookup of object cache based on their respective type
        self._index_caches: Dict[Type, Dict[Any, int]] = {
            EndReason: self._end_reasons,
            PoreType: self._pores,
            RunInfo: self._run_infos,
        }

        # Internal lookup of _add functions based on their respective type
        self._adder_funcs: Dict[Type, Callable[[Any], int]] = {
            EndReason: self._add_end_reason,
            PoreType: self._add_pore_type,
            RunInfo: self._add_run_info,
        }

    def __enter__(self) -> "Writer":
        return self

    def __exit__(self, *exc_details) -> None:
        self.close()

    def close(self) -> None:
        """Close the FileWriter handle"""
        safe_close(self, "_writer")
        self._writer = None

    @property
    def path(self) -> Path:
        """Return the path to the pod5 file"""
        return self._path

    @property
    def software_name(self) -> str:
        """Return the software name used to open this file"""
        return self._software_name

    def add(self, obj: Union[EndReason, PoreType, RunInfo]) -> int:
        """
        Add a :py:class:`EndReason`, :py:class:`PoreType`, or
        :py:class:`RunInfo` object to the Pod5 file (if it doesn't already
        exist) and return the index of this object in the Pod5 file.

        Parameters
        ----------
        obj : :py:class:`EndReason`, :py:class:`PoreType`, :py:class:`RunInfo`
            Object to find in this Pod5 file, adding it if it doesn't exist already

        Returns
        -------
        index : int
            The index of the object in the Pod5 file
        """
        # Get the index cache for the type of object given
        index_cache = self._index_caches[type(obj)]

        # Return the index of this object if it exists
        if obj in index_cache:
            return index_cache[obj]

        # Add object using the associated adder function e.g. _add_pore(pore: Pore)
        # and store the new index in the cache for future look-ups avoiding duplication
        added_index = self._adder_funcs[type(obj)](obj)
        index_cache[obj] = added_index

        # Return the newly added index
        return added_index

    def _add_end_reason(self, end_reason: EndReason) -> int:
        """Add the given EndReason instance to the pod5 file returning its index"""
        if self._writer is None:
            raise Pod5ApiException("Writer handle has been closed")
        return self._writer.add_end_reason(end_reason.reason.value)

    def _add_pore_type(self, pore_type: PoreType) -> int:
        """Add the given PoreType instance to the pod5 file returning its index"""
        if self._writer is None:
            raise Pod5ApiException("Writer handle has been closed")
        return self._writer.add_pore(pore_type)

    def _add_run_info(self, run_info: RunInfo) -> int:
        """Add the given RunInfo instance to the pod5 file returning its index"""
        if self._writer is None:
            raise Pod5ApiException("Writer handle has been closed")

        return self._writer.add_run_info(
            run_info.acquisition_id,
            timestamp_to_int(run_info.acquisition_start_time),
            run_info.adc_max,
            run_info.adc_min,
            map_to_tuples(run_info.context_tags),
            run_info.experiment_name,
            run_info.flow_cell_id,
            run_info.flow_cell_product_code,
            run_info.protocol_name,
            run_info.protocol_run_id,
            timestamp_to_int(run_info.protocol_start_time),
            run_info.sample_id,
            run_info.sample_rate,
            run_info.sequencing_kit,
            run_info.sequencer_position,
            run_info.sequencer_position_type,
            run_info.software,
            run_info.system_name,
            run_info.system_type,
            map_to_tuples(run_info.tracking_id),
        )

    def contains(self, obj: Union[EndReason, RunInfo]) -> bool:
        """
        Test if this Pod5 file contains the given object.

        Parameters
        ----------
        obj: :py:class:`EndReason`, :py:class:`RunInfo`
            Object to find in this Pod5 file

        Returns
        -------
        True if obj has already been added to this file
        """
        return obj in self._index_caches[type(obj)]

    def find(self, obj: Union[EndReason, RunInfo]) -> int:
        """
        Returns the index of obj in this Pod5 file raising a KeyError if it is missing.

        Parameters
        ----------
        obj: :py:class:`EndReason`, :py:class:`RunInfo`
            Obj instance to find in this Pod5 file

        Returns
        -------
        The index of the object in this Pod5 file

        Raises
        ------
        KeyError
            If the object is not in this file
        """
        try:
            return self._index_caches[type(obj)][obj]
        except KeyError as exc:
            raise KeyError(
                f"Could not find index of {obj} in Pod5 file writer: {self}"
            ) from exc

    def add_read(self, read: Union[Read, CompressedRead]) -> None:
        """
        Add a record to the open POD5 file with either compressed or uncompressed
        signal data depending on the given type of Read.

        Parameters
        ----------
        read : :py:class:`Read`, :py:class:`CompressedRead`
            POD5 Read or CompressedRead object to add as a record to the POD5 file.
        """
        self.add_reads([read])

    def add_reads(self, reads: Sequence[Union[Read, CompressedRead]]) -> None:
        """
        Add Read objects (with uncompressed signal data) as records in the open POD5
        file.

        Parameters
        ----------
        reads : Sequence of :py:class:`Read` or :py:class:`CompressedRead` exclusively
            List of Read object to be added to this POD5 file
        """

        # Nothing to do
        if not reads:
            return

        if self._writer is None:
            raise Pod5ApiException("Writer handle has been closed")

        if isinstance(reads[0], Read):
            return self._writer.add_reads(  # type: ignore [call-arg]
                *self._prepare_add_reads_args(reads),
                [r.signal for r in reads],  # type: ignore
            )
        elif isinstance(reads[0], CompressedRead):
            signal_chunks = [r.signal_chunks for r in reads]  # type: ignore
            signal_chunk_lengths = [r.signal_chunk_lengths for r in reads]  # type: ignore

            # Array containing the number of chunks for each signal
            signal_chunk_counts = np.array(
                [len(samples_per_chunk) for samples_per_chunk in signal_chunk_lengths],
                dtype=np.uint32,
            )

            return self._writer.add_reads_pre_compressed(  # type: ignore [call-arg]
                *self._prepare_add_reads_args(reads),
                # Join all signal data into one list
                list(itertools.chain(*signal_chunks)),
                # Join all read sample counts into one array
                np.concatenate(signal_chunk_lengths).astype(np.uint32),  # type: ignore [no-untyped-call]
                signal_chunk_counts,
            )

    def _prepare_add_reads_args(self, reads: Sequence[BaseRead]) -> List[Any]:
        """
        Converts the List of reads into the list of ctypes arrays of data to be supplied
        to the c api.
        """
        read_id = np.array(
            [np.frombuffer(read.read_id.bytes, dtype=np.uint8) for read in reads]
        )
        read_number = np.array([read.read_number for read in reads], dtype=np.uint32)
        start_sample = np.array([read.start_sample for read in reads], dtype=np.uint64)
        channel = np.array([read.pore.channel for read in reads], dtype=np.uint16)
        well = np.array([read.pore.well for read in reads], dtype=np.uint8)
        pore_type = np.array(
            [self.add(PoreType(read.pore.pore_type)) for read in reads], dtype=np.int16
        )
        calib_offset = np.array(
            [read.calibration.offset for read in reads], dtype=np.float32
        )
        calib_scale = np.array(
            [read.calibration.scale for read in reads], dtype=np.float32
        )
        median_before = np.array(
            [read.median_before for read in reads], dtype=np.float32
        )
        end_reason = np.array(
            [self.add(read.end_reason) for read in reads], dtype=np.int16
        )
        end_reason_forced = np.array(
            [read.end_reason.forced for read in reads], dtype=np.bool_
        )
        run_info = np.array([self.add(read.run_info) for read in reads], dtype=np.int16)
        num_minknow_events = np.array(
            [read.num_minknow_events for read in reads], dtype=np.uint64
        )
        tracked_scaling_scale = np.array(
            [read.tracked_scaling.scale for read in reads], dtype=np.float32
        )
        tracked_scaling_shift = np.array(
            [read.tracked_scaling.shift for read in reads], dtype=np.float32
        )
        predicted_scaling_scale = np.array(
            [read.predicted_scaling.scale for read in reads], dtype=np.float32
        )
        predicted_scaling_shift = np.array(
            [read.predicted_scaling.shift for read in reads], dtype=np.float32
        )
        num_reads_since_mux_change = np.array(
            [read.num_reads_since_mux_change for read in reads], dtype=np.uint32
        )
        time_since_mux_change = np.array(
            [read.time_since_mux_change for read in reads], dtype=np.float32
        )

        return [
            read_id.shape[0],
            read_id,
            read_number,
            start_sample,
            channel,
            well,
            pore_type,
            calib_offset,
            calib_scale,
            median_before,
            end_reason,
            end_reason_forced,
            run_info,
            num_minknow_events,
            tracked_scaling_scale,
            tracked_scaling_shift,
            predicted_scaling_scale,
            predicted_scaling_shift,
            num_reads_since_mux_change,
            time_since_mux_change,
        ]



from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache, partial
import os
from pathlib import Path
from typing import (
    Callable,
    Collection,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Union,
)
import warnings
from pod5.api_utils import Pod5ApiException

from pod5.pod5_types import PathOrStr
from pod5.reader import ReadRecord, Reader
from pod5.tools.utils import search_path

DEFAULT_CPUS = os.cpu_count() or 1


class DatasetReader:
    def __init__(
        self,
        paths: Union[PathOrStr, Collection[PathOrStr]],
        recursive: bool = False,
        pattern: str = "*.pod5",
        index: bool = False,
        threads: int = DEFAULT_CPUS,
        max_cached_readers: Optional[int] = 2**4,
        warn_duplicate_indexing: bool = True,
    ) -> None:
        """
        Reads pod5 files and/or directories of pod5 files as a dataset.

        Parameters
        ----------
        paths : PathOrStr | Collection[PathOrStr]
            One or more files or directories to load
        recursive : bool
            Search directories in `paths` recursively
        pattern : str
            A glob expression to match against file names
        index : bool
            Promptly index the dataset instead of deferring until required
        threads : int
            The number of threads to use
        max_cached_readers :  Optional[int]
            The maximum size of the `Reader` LRU cache. Set to `None` for an unlimited
            cache size.
        warn_duplicate_indexing : bool
            Issue warnings when duplicate read_ids are detected and
            indexing by read_id is attempted

        Note
        ----
        Random record access is implemented by creating an index of read_id to file
        path. This can consume a large amount of memory. Methods that generate an index
        have this noted in their docstring.

        Warnings
        --------
        If duplicate read_ids are present in the dataset, iterator methods such
        as `reads()` will yield all copies. Indexing methods such as `get_read`
        return one chosen randomly and issue a warning which can be suppressed by
        setting `warn_duplicate_indexing=False`
        """
        self._paths: List[Path] = sorted(
            self._collect_dataset(
                paths, recursive=recursive, pattern=pattern, threads=threads
            )
        )
        self._num_reads: Optional[int] = None
        self._max_cached_readers = max_cached_readers
        self.threads = threads
        self.warn_duplicate_indexing = warn_duplicate_indexing

        # Cache on DatasetReader instances and control cache size on init
        self._get_reader = self._init_get_reader(self._max_cached_readers)

        if index:
            self._index_read_ids()
        else:
            self._index: Optional[Dict[str, Path]] = None

    def __iter__(self) -> Generator[ReadRecord, None, None]:
        yield from self.reads()

    def __len__(self) -> int:
        """Returns the number of reads in this dataset"""
        return self.num_reads

    @property
    def num_reads(self) -> int:
        """
        Return the number of `ReadRecords` in this dataset.
        """
        if self._num_reads is not None:
            return self._num_reads

        def _get_num_reads(path: Path) -> int:
            try:
                return self.get_reader(path).num_reads
            except Exception as exc:
                msg = f"DatasetReader error reading: {[path]}"
                raise Pod5ApiException(msg) from exc

        self._num_reads = 0
        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            self._num_reads = sum(executor.map(_get_num_reads, self._paths))

        return self._num_reads

    @property
    def paths(self) -> List[Path]:
        """Return the list of pod5 file paths in this dataset"""
        return self._paths

    @property
    def read_ids(self) -> Generator[str, None, None]:
        """
        Yield all read_ids in this dataset
        """

        def _get_read_ids(path: Path):
            return self.get_reader(path).read_ids

        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            for read_id_gen in executor.map(_get_read_ids, self.paths):
                yield from read_id_gen

    def reads(
        self,
        selection: Optional[Iterable[str]] = None,
        preload: Optional[Set[str]] = None,
    ) -> Generator[ReadRecord, None, None]:
        """
        Iterate over ``ReadRecord``s in the dataset.

        Parameters
        ----------
        selection : iterable[str]
            The read ids to walk in the file.
        preload : set[str]
            Columns to preload - "samples" and "sample_count" are valid values

        Note
        ----
        ``ReadRecord``s are yielded in on-disk record order for each file in ``self.paths``.

        Missing records are not detected and multiple records will be
        yielded if there are duplicates in either of the dataset or selection.

        Yields
        ------
        :py:class:`ReadRecord`
        """

        def _get_reads_iter(path: Path) -> Generator[ReadRecord, None, None]:
            return self.get_reader(path).reads(
                selection=selection, missing_ok=True, preload=preload
            )

        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            for read_gen in executor.map(_get_reads_iter, self.paths):
                yield from read_gen

    def get_read(self, read_id: str) -> Optional[ReadRecord]:
        """
        Get a `ReadRecord` by `read_id` or return `None` if it is missing

        Parameters
        ----------
        read_id : str
            The read_id (UUID) string in this dataset to find

        Note
        ----
        This method will index the dataset

        Warnings
        --------
        Issues a warning if duplicate read_ids are detected in this dataset.
        The returned `ReadRecord` is a always valid but the source may be random
        between instances of a `DatasetReader`.

        Returns
        -------
        A :py:class:`ReadRecord` or `None`
        """
        path = self.get_path(read_id)
        if path is None:
            return None

        reader = self.get_reader(path)
        try:
            return next(reader.reads(selection=[read_id]))
        except StopIteration:
            return None

    @staticmethod
    def _init_get_reader(maxsize: Optional[int]) -> Callable[[Path], Reader]:
        # This wrapper allows the size of the LRU cache to be set during initialization
        # without global variables
        @lru_cache(maxsize=maxsize)
        def _get_reader(path: Path) -> Reader:
            return Reader(path)

        return _get_reader

    def get_reader(self, path: PathOrStr) -> Reader:
        """
        Get a pod5 file `Reader` in this dataset by `path`

        Parameters
        ----------
        path : PathOrStr
            Path to a pod5 file

        Returns
        -------
        A :py:class:`Reader`
        """
        return self._get_reader(Path(path))

    def get_path(self, read_id: str) -> Optional[Path]:
        """
        Get the pod5 `Path` for a given `read_id` or `None` if it was not found

        Parameters
        ----------
        read_id : str
            The read_id (UUID) string in this dataset

        Note
        ----
        This method will index the dataset

        Warnings
        --------
        Issues a warning if duplicate read_ids are detected in this dataset.
        The returned path is a always valid file which contains this read_id but this
        may be random between instances.

        Returns
        -------
        A `Path` or `None`
        """
        if self._index is None:
            self._index_read_ids()

        if self._index is None:
            return None

        if self.has_duplicate():
            self._issue_duplicate_read_warning()

        return self._index.get(read_id, None)

    def clear_readers(self) -> None:
        """Clears the readers LRU cache"""
        self._get_reader.cache_clear()  # type: ignore

    def clear_index(self) -> None:
        """Clears the read_id to file path index"""
        self._index = None

    def has_duplicate(self) -> bool:
        """
        Returns `True` if there are duplicate `read_ids` in this dataset

        Note
        ----
        This method will index the dataset
        """
        if self._index is None:
            self._index_read_ids()
        assert self._index is not None
        return len(self) != len(self._index)

    @staticmethod
    def _collect_dataset(
        paths: Union[PathOrStr, Collection[PathOrStr]],
        recursive: bool,
        pattern: str,
        threads: int,
    ) -> Set[Path]:
        if isinstance(paths, (str, Path, os.PathLike)):
            paths = [Path(paths)]

        _paths = list(paths)
        collected: Set[Path] = set()
        with ThreadPoolExecutor(max_workers=threads) as executor:
            search = partial(search_path, recursive=recursive, patterns=[pattern])
            for coll in executor.map(search, _paths):
                collected.update(coll)
        return collected

    def _index_read_ids(self) -> None:
        self._index = {}

        def _get_index(path: Path) -> Dict[str, Path]:
            try:
                return {read_id: path for read_id in self.get_reader(path).read_ids}
            except Exception as exc:
                msg = f"DatasetReader error reading: {[path]}"
                raise Pod5ApiException(msg) from exc

        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            for index in executor.map(_get_index, self.paths):
                self._index.update(index)

    def _issue_duplicate_read_warning(self) -> None:
        if self.warn_duplicate_indexing:
            warnings.warn("duplicate read_ids found in dataset")

    def __enter__(self) -> "DatasetReader":
        return self

    def __exit__(self, *exc_details) -> None:
        self.clear_index()
        self.clear_readers()



"""
Tool for converting pod5 files to the legacy fast5 format
"""
import time
from concurrent.futures import Future, ProcessPoolExecutor, as_completed
from pathlib import Path
from typing import Dict, List, Tuple

import h5py
import numpy
import vbz_h5py_plugin  # noqa: F401
from more_itertools import chunked

import pod5 as p5
from pod5.tools.parsers import pod5_convert_to_fast5_argparser, run_tool
from pod5.tools.utils import DEFAULT_THREADS, collect_inputs, limit_threads

# Pod5 does not have 'partial' so need to add that back in here.
FAST5_END_REASONS = {
    "unknown": 0,
    "partial": 1,  # Do not remove, required by fast5.
    "mux_change": 2,
    "unblock_mux_change": 3,
    "data_service_unblock_mux_change": 4,
    "signal_positive": 5,
    "signal_negative": 6,
}

# Fast5 types
FAST5_END_REASON_TYPE = h5py.enum_dtype(FAST5_END_REASONS)
FAST5_STRING_TYPE = h5py.string_dtype("ascii")


class StatusMonitor:
    """Class for monitoring the status / progress of the conversion"""

    def __init__(self, file_count: int):
        self.update_interval = 10

        self.file_count = file_count
        self.files_started = 0
        self.files_ended = 0
        self.read_count = 0
        self.reads_processed = 0
        self.sample_count = 0

        self.time_start = self.time_last_update = time.time()

    @property
    def running(self) -> bool:
        """Return true if not all files have finished processing"""
        return self.files_ended < self.file_count

    def increment(
        self,
        *,
        files_started: int = 0,
        files_ended: int = 0,
        read_count: int = 0,
        reads_processed: int = 0,
        sample_count: int = 0,
    ) -> None:
        """Incremeent the status counters"""
        self.files_started += files_started
        self.files_ended += files_ended
        self.read_count += read_count
        self.reads_processed += reads_processed
        self.sample_count += sample_count

    @property
    def samples_mb(self) -> float:
        """Return the samples count in megabytes"""
        return (self.sample_count * 2) / 1_000_000

    @property
    def time_elapsed(self) -> float:
        """Return the total time elapsed in seconds"""
        return self.time_last_update - self.time_start

    @property
    def sample_rate(self) -> float:
        """Return the time averaged sample rate"""
        return self.samples_mb / self.time_elapsed

    def print_status(self, force: bool = False):
        """Print the status if the update interval has passed or if forced"""
        now = time.time()

        if force or self.time_last_update + self.update_interval < now:
            self.time_last_update = now

            print(
                f"{self.reads_processed} reads,\t",
                f"{self.formatted_sample_count},\t",
                f"{self.files_ended}/{self.file_count} files,\t",
                f"{self.sample_rate:.1f} MB/s",
            )

    @property
    def formatted_sample_count(self) -> str:
        """Return the sample count as a string with leading Metric prefix if necessary"""
        units = [
            (1000000000000, "T"),
            (1000000000, "G"),
            (1000000, "M"),
            (1000, "K"),
        ]

        for div, unit in units:
            if self.sample_count > div:
                return f"{self.sample_count/div:.1f} {unit}Samples"
        return f"{self.sample_count} Samples"


def write_pod5_record_to_fast5(read: p5.ReadRecord, fast5: h5py.File) -> None:
    tracking_id = read.run_info.tracking_id

    read_group = fast5.create_group(f"read_{read.read_id}")
    read_group.attrs.create(
        "run_id",
        tracking_id["run_id"].encode("ascii"),
        dtype=FAST5_STRING_TYPE,
    )
    read_group.attrs.create(
        "pore_type",
        read.pore.pore_type.encode("ascii"),
        dtype=FAST5_STRING_TYPE,
    )

    tracking_id_group = read_group.create_group("tracking_id")
    for k, v in tracking_id.items():
        tracking_id_group.attrs[k] = v

    context_tags_group = read_group.create_group("context_tags")
    for k, v in read.run_info.context_tags.items():
        context_tags_group.attrs[k] = v

    channel_id_group = read_group.create_group("channel_id")
    digitisation = read.run_info.adc_max - read.run_info.adc_min + 1
    channel_id_group.attrs.create("digitisation", digitisation, dtype=numpy.float64)
    channel_id_group.attrs.create(
        "offset", read.calibration.offset, dtype=numpy.float64
    )

    channel_id_group.attrs.create(
        "range", digitisation * read.calibration.scale, dtype=numpy.float64
    )
    channel_id_group.attrs.create(
        "sampling_rate", read.run_info.sample_rate, dtype=numpy.float64
    )
    channel_id_group.attrs["channel_number"] = str(read.pore.channel)

    raw_group = read_group.create_group("Raw")
    raw_group.create_dataset(
        "Signal",
        data=read.signal,
        dtype=numpy.int16,
        compression=32020,
        compression_opts=(0, 2, 1, 1),
    )
    raw_group.attrs.create("start_time", read.start_sample, dtype=numpy.uint64)
    raw_group.attrs.create("duration", read.sample_count, dtype=numpy.uint32)
    raw_group.attrs.create("read_number", read.read_number, dtype=numpy.int32)
    raw_group.attrs.create("start_mux", read.pore.well, dtype=numpy.uint8)
    raw_group.attrs["read_id"] = str(read.read_id).encode("utf-8")
    raw_group.attrs.create("median_before", read.median_before, dtype=numpy.float64)

    # Lookup the fast5 enumeration values, which should include "partial: 1"
    # This will ensure that the enumeration is valid on a round-trip
    raw_group.attrs.create(
        "end_reason",
        FAST5_END_REASONS[read.end_reason.name],
        dtype=FAST5_END_REASON_TYPE,
    )

    raw_group.attrs.create(
        "num_minknow_events", read.num_minknow_events, dtype=numpy.uint64
    )

    raw_group.attrs.create(
        "tracked_scaling_scale",
        read.tracked_scaling.scale,
        dtype=numpy.float32,
    )
    raw_group.attrs.create(
        "tracked_scaling_shift",
        read.tracked_scaling.shift,
        dtype=numpy.float32,
    )
    raw_group.attrs.create(
        "predicted_scaling_scale",
        read.predicted_scaling.scale,
        dtype=numpy.float32,
    )
    raw_group.attrs.create(
        "predicted_scaling_shift",
        read.predicted_scaling.shift,
        dtype=numpy.float32,
    )
    raw_group.attrs.create(
        "num_reads_since_mux_change",
        read.num_reads_since_mux_change,
        dtype=numpy.uint32,
    )
    raw_group.attrs.create(
        "time_since_mux_change",
        read.time_since_mux_change,
        dtype=numpy.float32,
    )


def convert_pod5_to_fast5(
    source: Path, dest: Path, read_ids: List[str]
) -> Tuple[int, int]:
    """
    Open a source pod5 file and write the selected read_ids into the destination fast5
    file target.
    """

    dest.parent.mkdir(parents=True, exist_ok=True)

    if dest.exists():
        dest.unlink()

    total_samples = 0

    with p5.Reader(source) as reader:
        with h5py.File(dest, "w") as f5:
            f5.attrs.create(
                "file_version", "3.0".encode("ascii"), dtype=FAST5_STRING_TYPE
            )
            f5.attrs.create(
                "file_type", "multi-read".encode("ascii"), dtype=FAST5_STRING_TYPE
            )

            # Take the chunk of read ids for this file
            for read in reader.reads(
                selection=read_ids, missing_ok=False, preload={"samples"}
            ):
                write_pod5_record_to_fast5(read, f5)

                total_samples += read.num_samples

    return (len(read_ids), total_samples)


def convert_to_fast5(
    inputs: List[Path],
    output: Path,
    recursive: bool = False,
    threads: int = DEFAULT_THREADS,
    force_overwrite: bool = False,
    file_read_count: int = 4000,
):
    if output.exists() and not output.is_dir():
        raise FileExistsError("Cannot output to a file")

    threads = limit_threads(threads)

    with ProcessPoolExecutor(max_workers=threads) as executor:
        total_reads = 0
        futures: Dict[Future, Path] = {}

        # Enumerate over input pod5 files
        for input_idx, source in enumerate(
            collect_inputs(inputs, recursive, "*.pod5", threads=threads)
        ):
            # Open the inputs to read the read ids
            with p5.Reader(source) as reader:
                for chunk_idx, read_ids in enumerate(
                    chunked(reader.read_ids, file_read_count)
                ):
                    dest = (
                        output / f"{source.stem}.{chunk_idx}_{input_idx}.fast5"
                    ).resolve()

                    if dest.exists() and not force_overwrite:
                        raise FileExistsError(
                            "Output path points to an existing file and --force-overwrite not set"
                        )

                    kwargs = {
                        "source": source,
                        "dest": dest,
                        "read_ids": read_ids,
                    }
                    futures[executor.submit(convert_pod5_to_fast5, **kwargs)] = dest  # type: ignore

                total_reads += len(reader.read_ids)

        print(f"Converting pod5s into {len(futures)} fast5 files. Please wait...")

        status = StatusMonitor(file_count=len(inputs))
        status.increment(files_started=len(inputs), read_count=total_reads)

        for idx, future in enumerate(as_completed(futures)):
            (reads_converted, samples_converted) = future.result()

            status.increment(
                files_ended=1,
                sample_count=samples_converted,
                reads_processed=reads_converted,
            )
            status.print_status()

        status.print_status(force=True)

        print("Conversion complete")


def main():
    run_tool(pod5_convert_to_fast5_argparser())


if __name__ == "__main__":
    main()



"""
Tool for converting fast5 files to the pod5 format
"""

import datetime
import multiprocessing as mp
from multiprocessing.context import SpawnContext
import sys
import warnings
from pod5.pod5_types import CompressedRead
from tqdm.auto import tqdm
import uuid
from pathlib import Path
from queue import Empty
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import h5py
import iso8601
import more_itertools
import vbz_h5py_plugin  # noqa: F401

import pod5 as p5
from pod5.signal_tools import DEFAULT_SIGNAL_CHUNK_SIZE, vbz_compress_signal_chunked
from pod5.tools.parsers import pod5_convert_from_fast5_argparser, run_tool
from pod5.tools.utils import (
    DEFAULT_THREADS,
    PBAR_DEFAULTS,
    collect_inputs,
    init_logging,
    limit_threads,
    logged,
    logged_all,
    terminate_processes,
)

READ_CHUNK_SIZE = 400
TIMEOUT_SECONDS = 600


logger = init_logging()


class QueueManager:
    def __init__(
        self,
        context: SpawnContext,
        inputs: Collection[Path],
        threads: int,
        timeout: float,
    ) -> None:
        """Manager for balancing work queues"""
        self._requests_size = threads * 2
        self._inputs: mp.Queue = context.Queue(maxsize=len(inputs))
        self._requests: mp.Queue = context.Queue(maxsize=self._requests_size)
        self._data: mp.Queue = context.Queue()
        self._exceptions: mp.Queue = context.Queue()
        self._timeout = timeout

        self._start(inputs=inputs)

    def _await(self, queue: mp.Queue) -> Any:
        """Await the next item on a queue raising TimeoutError if failing"""
        try:
            item = queue.get(timeout=self._timeout)
            return item
        except Empty:
            logger.fatal("Empty queue or timeout ")
            raise TimeoutError(f"No progress in {self._timeout} seconds - quitting")

    def enqueue_request(self) -> None:
        self._requests.put(None, timeout=self._timeout)

    def await_request(self) -> None:
        """Await a request for data"""
        self._await(self._requests)

    @logged()
    def enqueue_data(
        self, path: Optional[Path], reads: Union[List[CompressedRead], int, None]
    ) -> None:
        """
        Enqueues an input path and either a list of compressed reads to be written, or
        the total count of reads converted for that path.
        Otherwise, if path is None, mark the child process as being empty.
        """
        self._data.put((path, reads), timeout=self._timeout)

    @logged(log_time=True)
    def await_data(
        self,
    ) -> Tuple[Optional[Path], Union[List[CompressedRead], int, None]]:
        """
        Await compressed reads or the total count of reads compressed (file end) for
        a input filepath. Enqueues the next request if necessary
        """
        path, item = self._await(self._data)

        # Check for the exhausted process sentinel value
        if path is None:
            return None, None

        # Add another request if we received compressed reads
        if isinstance(item, List):
            self.enqueue_request()

        return path, item

    @logged(log_args=True)
    def enqueue_exception(self, path: Path, exception: Exception, trace: str) -> None:
        self._exceptions.put((path, exception, trace), timeout=self._timeout)

    def get_exception(self) -> Optional[Tuple[Path, Exception, str]]:
        """Promptly get an exception if any"""
        try:
            # Use short timeout instead of get_nowait as we might call this method
            # very shortly after enqueueing an exception
            path, exc, trace = self._exceptions.get(timeout=0.01)
            logger.exception(f"Encountered an exception in {path} - {exc}")
            if trace:
                logger.exception(f"Trace Exception {path}\n{trace}")
            return path, exc, trace
        except Empty:
            pass
        return None

    @logged(log_args=True)
    def enqueue_input(self, path: Path) -> None:
        """Enqueue a request"""
        self._inputs.put(path)

    @logged_all
    def get_input(self) -> Optional[Path]:
        """Promptly get an input if any returning None if queue is empty"""
        try:
            return self._inputs.get(timeout=0.1)
        except Empty:
            pass
        return None

    @logged(log_return=True)
    def _discard_and_close(self, queue: mp.Queue) -> int:
        """
        Discard all remaining enqueued items and close a queue to nicely shutdown the
        queue. Returns the number of discarded items
        """
        count = 0
        while True:
            try:
                queue.get(timeout=0.1)
                count += 1
            except Exception:
                break
        queue.close()
        queue.join_thread()
        return count

    @logged(log_return=True)
    def shutdown(self) -> Tuple[int, int, int, int]:
        """Shutdown all queues returning the counts of all remaining items"""
        n_inputs = self._discard_and_close(self._inputs)
        n_req = self._discard_and_close(self._requests)
        n_data = self._discard_and_close(self._data)
        n_exc = self._discard_and_close(self._exceptions)

        if n_inputs > 0:
            logger.warn("Unfinished inputs found during shutdown!")
        if n_data > 0:
            logger.warn("Unfinished data found during shutdown!")
        if n_exc > 0:
            logger.warn("Unfinished exceptions found during shutdown!")

        return n_inputs, n_req, n_data, n_exc

    @logged(log_args=True)
    def _start(self, inputs: Iterable[Path]) -> None:
        """Enqueue all inputs for child processes to poll and set the requests size"""
        for path in inputs:
            if path.is_file():
                self.enqueue_input(path)

        for _ in range(self._requests_size):
            self.enqueue_request()


class OutputHandler:
    """Class for managing p5.Writer handles"""

    @logged(log_args=True)
    def __init__(
        self,
        output_root: Path,
        one_to_one: Optional[Path],
        force_overwrite: bool,
    ):
        self.output_root = output_root
        self._one_to_one = one_to_one
        self._force_overwrite = force_overwrite
        self._input_to_output: Dict[Path, Path] = {}
        self._open_writers: Dict[Path, p5.Writer] = {}
        self._closed_writers: Dict[Path, bool] = {}

    @logged_all
    def _open_writer(self, output_path: Path) -> Optional[p5.Writer]:
        """Get the writer from existing handles or create a new one if unseen"""
        if output_path in self._open_writers:
            return self._open_writers[output_path]

        if output_path in self._closed_writers:
            had_exception = self._closed_writers[output_path]
            if had_exception:
                return None
            raise FileExistsError(f"Trying to re-open a closed Writer to {output_path}")

        if output_path.exists() and self._force_overwrite:
            output_path.unlink()

        writer = p5.Writer(output_path)
        self._open_writers[output_path] = writer
        return writer

    @logged_all
    def get_writer(self, input_path: Path) -> Optional[p5.Writer]:
        """Get a Pod5Writer to write data from the input_path"""
        if input_path not in self._input_to_output:
            out_path = self.resolve_output_path(
                path=input_path, root=self.output_root, relative_root=self._one_to_one
            )
            self._input_to_output[input_path] = out_path

        output_path = self._input_to_output[input_path]
        return self._open_writer(output_path=output_path)

    @staticmethod
    @logged_all
    def resolve_one_to_one_path(path: Path, root: Path, relative_root: Path):
        """
        Find the relative path between the input path and the relative root
        """
        try:
            relative = path.with_suffix(".pod5").relative_to(relative_root)
        except ValueError as exc:
            raise RuntimeError(
                f"--one-to-one directory must be a relative parent of "
                f"all input fast5 files. For {path} relative to {relative_root}"
            ) from exc

        # Resolve the new final output path relative to the output directory
        # This path is to a file with the equivalent filename(.pod5)
        return root / relative

    @staticmethod
    @logged_all
    def resolve_output_path(
        path: Path, root: Path, relative_root: Optional[Path]
    ) -> Path:
        """
        Resolve the output path. If relative_root is a path, resolve the relative output
        path under root, otherwise, the output is either root or a new file within root
        if root is a directory
        """
        if relative_root is not None:
            # Resolve the relative path to the one_to_one root path
            out_path = OutputHandler.resolve_one_to_one_path(
                path=path,
                root=root,
                relative_root=relative_root,
            )

            # Create directory structure if needed
            out_path.parent.mkdir(parents=True, exist_ok=True)
            return out_path

        if root.is_dir():
            # If the output path is a directory, the write the default filename
            return root / "output.pod5"

        # The provided output path is assumed to be a named file
        return root

    @logged(log_args=True)
    def set_input_complete(self, input_path: Path, is_exception: bool) -> None:
        """Close the Pod5Writer for associated input_path"""
        if not self._one_to_one:
            # Do not close common output file when not in 1-2-1 mode
            return

        if input_path not in self._input_to_output:
            return

        output_path = self._input_to_output[input_path]
        self._open_writers[output_path].close()
        self._closed_writers[output_path] = is_exception
        del self._open_writers[output_path]

    @logged()
    def close_all(self):
        """Close all open writers"""
        for path, writer in self._open_writers.items():
            writer.close()
            del writer
            # Keep track of closed writers to ensure we don't overwrite our own work
            self._closed_writers[path] = False
        self._open_writers = {}

    def __del__(self) -> None:
        self.close_all()


class StatusMonitor:
    """Class for monitoring the status of the conversion"""

    @logged_all
    def __init__(self, paths: Sequence[Path]):
        # Estimate that a fast5 file will have 4k reads
        self.path_reads = {path: 4000 for path in paths}
        self.count_finished = 0

        self.pbar = tqdm(
            total=self.total_reads,
            desc=f"Converting {len(self.path_reads)} Fast5s",
            unit="Reads",
            leave=True,
            **PBAR_DEFAULTS,
        )

    @property
    def total_files(self) -> int:
        return len(self.path_reads)

    @property
    def total_reads(self) -> int:
        return sum(self.path_reads.values())

    @logged(log_args=True)
    def increment_reads(self, n: int) -> None:
        """Increment the reads status by n"""
        self.pbar.update(n)

    @logged(log_args=True)
    def update_reads_total(self, path: Path, total: int) -> None:
        """Increment the reads status by n and update the total reads"""
        self.path_reads[path] = total
        self.pbar.total = self.total_reads
        self.pbar.refresh()

    @logged(log_args=True)
    def write(self, msg: str, file: Any) -> None:
        """Write runtime message to avoid clobbering tqdm pbar"""
        self.pbar.write(msg, file=file)

    @logged()
    def close(self) -> None:
        """Close the progress bar"""
        self.pbar.close()


@logged_all
def is_multi_read_fast5(path: Path) -> bool:
    """
    Assert that the given path points to a a multi-read fast5 file for which
    direct-to-pod5 conversion is supported.
    """
    try:
        with h5py.File(path) as _h5:
            # The "file_type" attribute might be present on supported multi-read fast5 files.
            if _h5.attrs.get("file_type") == "multi-read":
                return True

            # No keys, assume multi-read but there shouldn't be anything to do which would
            # cause an issue so pass silently
            if len(_h5) == 0:
                return True

            # if there are "read_x" keys, this is a multi-read file
            if any(key for key in _h5 if key.startswith("read_")):
                return True

    except Exception:
        pass

    return False


def decode_str(value: Union[str, bytes]) -> str:
    """Decode a h5py utf-8 byte string to python string"""
    if isinstance(value, str):
        return value
    return value.decode("utf-8")


def convert_fast5_end_reason(fast5_end_reason: int) -> p5.EndReason:
    """
    Return an EndReason instance from the given end_reason integer from a fast5 file.
    This will handle the difference between fast5 and pod5 values for this enumeration
    and set the default "forced" value for each fast5 enumeration value.
    """
    # Expected fast5 enumeration:
    # end_reason_dict = {
    #     "unknown": 0,
    #     "partial": 1, <-- Not used in pod5
    #     "mux_change": 2,  <-- Remaining values are offset by +1
    #     "unblock_mux_change": 3,
    #     "data_service_unblock_mux_change": 4,
    #     "signal_positive": 5,
    #     "signal_negative": 6,
    # }

    # (0:unknown | 1:partial) => pod5 (0:unknown)
    if fast5_end_reason < 2:
        return p5.EndReason.from_reason_with_default_forced(p5.EndReasonEnum.UNKNOWN)

    # Resolve the offset in enumeration values between both files
    p5_scaled_end_reason = fast5_end_reason - 1
    return p5.EndReason.from_reason_with_default_forced(
        p5.EndReasonEnum(p5_scaled_end_reason)
    )


def convert_datetime_as_epoch_ms(
    time_str: Union[str, bytes, None]
) -> datetime.datetime:
    """Convert the fast5 time string to timestamp"""
    epoch = datetime.datetime.utcfromtimestamp(0).replace(tzinfo=datetime.timezone.utc)
    if time_str is None:
        return epoch
    try:
        return iso8601.parse_date(decode_str(time_str))
    except iso8601.iso8601.ParseError:
        return epoch


def convert_run_info(
    acq_id: str,
    adc_max: int,
    adc_min: int,
    sample_rate: int,
    context_tags: Dict[str, Union[str, bytes]],
    device_type: str,
    tracking_id: Dict[str, Union[str, bytes]],
) -> p5.RunInfo:
    """Create a Pod5RunInfo instance from parsed fast5 data"""
    return p5.RunInfo(
        acquisition_id=acq_id,
        acquisition_start_time=convert_datetime_as_epoch_ms(
            tracking_id.get("exp_start_time")
        ),
        adc_max=adc_max,
        adc_min=adc_min,
        context_tags={
            str(key): decode_str(value) for key, value in context_tags.items()
        },
        experiment_name="",
        flow_cell_id=decode_str(tracking_id.get("flow_cell_id", b"")),
        flow_cell_product_code=decode_str(
            tracking_id.get("flow_cell_product_code", b"")
        ),
        protocol_name=decode_str(tracking_id.get("exp_script_name", b"")),
        protocol_run_id=decode_str(tracking_id.get("protocol_run_id", b"")),
        protocol_start_time=convert_datetime_as_epoch_ms(
            tracking_id.get("protocol_start_time", None)
        ),
        sample_id=decode_str(tracking_id.get("sample_id", b"")),
        sample_rate=sample_rate,
        sequencing_kit=decode_str(context_tags.get("sequencing_kit", b"")),
        sequencer_position=decode_str(tracking_id.get("device_id", b"")),
        sequencer_position_type=decode_str(tracking_id.get("device_type", device_type)),
        software="python-pod5-converter",
        system_name=decode_str(tracking_id.get("host_product_serial_number", b"")),
        system_type=decode_str(tracking_id.get("host_product_code", b"")),
        tracking_id={str(key): decode_str(value) for key, value in tracking_id.items()},
    )


def convert_fast5_read(
    fast5_read: h5py.Group,
    run_info_cache: Dict[str, p5.RunInfo],
    signal_chunk_size: int = DEFAULT_SIGNAL_CHUNK_SIZE,
) -> p5.CompressedRead:
    """
    Given a fast5 read parsed from a fast5 file, return a pod5.Read object.
    """
    channel_id = fast5_read["channel_id"]
    raw = fast5_read["Raw"]

    attrs = fast5_read.attrs

    # Get the acquisition id
    if "run_id" in attrs:
        acq_id = decode_str(attrs["run_id"])
    else:
        acq_id = decode_str(fast5_read["tracking_id"].attrs["run_id"])

    # Create new run_info if we've not seen this acquisition id before
    if acq_id not in run_info_cache:
        adc_min = 0
        adc_max = 2047
        device_type_guess = "promethion"
        if channel_id.attrs["digitisation"] == 8192:
            adc_min = -4096
            adc_max = 4095
            device_type_guess = "minion"

        # Add new run_info to cache
        run_info_cache[acq_id] = convert_run_info(
            acq_id=acq_id,
            adc_max=adc_max,
            adc_min=adc_min,
            sample_rate=int(channel_id.attrs["sampling_rate"]),
            context_tags=dict(fast5_read["context_tags"].attrs),
            device_type=device_type_guess,
            tracking_id=dict(fast5_read["tracking_id"].attrs),
        )

    # Process attributes unique to this read
    read_id = uuid.UUID(decode_str(raw.attrs["read_id"]))
    pore = p5.Pore(
        channel=int(channel_id.attrs["channel_number"]),
        well=raw.attrs["start_mux"],
        pore_type=decode_str(attrs.get("pore_type", b"not_set")),
    )
    calibration = p5.Calibration.from_range(
        offset=channel_id.attrs["offset"],
        adc_range=channel_id.attrs["range"],
        digitisation=channel_id.attrs["digitisation"],
    )

    end_reason = convert_fast5_end_reason(raw.attrs.get("end_reason", 0))

    # Signal conversion process
    signal = raw["Signal"][()]
    signal_chunks, signal_chunk_lengths = vbz_compress_signal_chunked(
        signal, signal_chunk_size
    )

    return p5.CompressedRead(
        read_id=read_id,
        pore=pore,
        calibration=calibration,
        read_number=raw.attrs["read_number"],
        start_sample=raw.attrs["start_time"],
        median_before=raw.attrs["median_before"],
        num_minknow_events=raw.attrs.get("num_minknow_events", 0),
        tracked_scaling=p5.pod5_types.ShiftScalePair(
            raw.attrs.get("tracked_scaling_shift", float("nan")),
            raw.attrs.get("tracked_scaling_scale", float("nan")),
        ),
        predicted_scaling=p5.pod5_types.ShiftScalePair(
            raw.attrs.get("predicted_scaling_shift", float("nan")),
            raw.attrs.get("predicted_scaling_scale", float("nan")),
        ),
        num_reads_since_mux_change=raw.attrs.get("num_reads_since_mux_change", 0),
        time_since_mux_change=raw.attrs.get("time_since_mux_change", 0.0),
        end_reason=end_reason,
        run_info=run_info_cache[acq_id],
        signal_chunks=signal_chunks,
        signal_chunk_lengths=signal_chunk_lengths,
    )


def get_read_from_fast5(group_name: str, h5_file: h5py.File) -> Optional[h5py.Group]:
    """Read a group from a h5 file ensuring that it's a read"""
    if not group_name.startswith("read_"):
        return None

    try:
        return h5_file[group_name]
    except KeyError as exc:
        # Observed strange behaviour where h5py reports a KeyError with
        # the message "Unable to open object". Report a failed read as warning
        warnings.warn(
            f"Failed to read key {group_name} from {h5_file.filename} : {exc}",
        )
    return None


def convert_fast5_file_chunk(
    queues: QueueManager,
    handle: h5py.File,
    chunk: Iterable[str],
    cache: Dict[str, p5.RunInfo],
    signal_chunk_size: int,
) -> List[CompressedRead]:
    reads: List[p5.CompressedRead] = []

    # Allow request queue to throttle work
    queues.await_request()
    try:
        for group_name in chunk:
            f5_read = get_read_from_fast5(group_name, handle)
            if f5_read is None:
                continue
            read = convert_fast5_read(f5_read, cache, signal_chunk_size)
            reads.append(read)

    except Exception as exc:
        # Ensures that requests aren't exhausted
        queues.enqueue_request()
        raise exc
    return reads


@logged_all
def convert_fast5_file(
    path: Path,
    queues: QueueManager,
    signal_chunk_size: int = DEFAULT_SIGNAL_CHUNK_SIZE,
) -> int:
    """Convert the reads in a fast5 file"""

    run_info_cache: Dict[str, p5.RunInfo] = {}
    total_reads: int = 0

    with h5py.File(str(path), "r") as _f5:
        for chunk in more_itertools.chunked(_f5.keys(), READ_CHUNK_SIZE):
            reads = convert_fast5_file_chunk(
                queues, _f5, chunk, run_info_cache, signal_chunk_size
            )
            queues.enqueue_data(path, reads)
            total_reads += len(reads)

    return total_reads


@logged()
def issue_not_multi_read_exception(path: Path, queues: QueueManager):
    logger.error(f"Input {path.name} is not a multi-read fast5")
    queues.enqueue_exception(
        path=path,
        exception=TypeError(f"{path} is not a multi-read fast5 file."),
        trace="",
    )
    logger.info(f"Enqueueing file end: {path.name} reads: 0")
    queues.enqueue_data(path, 0)


@logged(log_time=True)
def convert_fast5_files(
    queues: QueueManager,
    signal_chunk_size: int = DEFAULT_SIGNAL_CHUNK_SIZE,
) -> None:
    """
    Main function for converting fast5s available in queues.
    Collections of converted reads are emplaced on the data_queue for writing in
    the main process.
    """
    while True:
        path = queues.get_input()

        if path is None:
            logger.info("Inputs exhausted. Closing Process")
            break

        if not is_multi_read_fast5(path):
            issue_not_multi_read_exception(path, queues)
            continue

        try:
            total_reads = convert_fast5_file(path, queues, signal_chunk_size)
            logger.info(f"Enqueueing file end: {path.name} reads: {total_reads}")
            queues.enqueue_data(path, total_reads)

        except Exception as exc:
            import traceback

            logger.error(f"Enqueueing exception: {path.name} {exc}")
            queues.enqueue_exception(path, exc, traceback.format_exc())

    logger.info("Enqueue sentinel")
    queues.enqueue_data(None, None)


@logged(log_args=True)
def handle_exception(
    exception: Tuple[Path, Exception, str],
    output_handler: OutputHandler,
    status: StatusMonitor,
    strict: bool,
) -> None:
    path, exc, trace = exception
    status.write(str(exc), sys.stderr)
    output_handler.set_input_complete(path, is_exception=True)

    if strict:
        status.close()
        logger.fatal("Exception raised and --strict set")
        logger.debug(f"trace: {trace}")
        raise exc


@logged_all
def process_conversion_tasks(
    queues: QueueManager,
    output_handler: OutputHandler,
    status: StatusMonitor,
    strict: bool,
    threads: int = DEFAULT_THREADS,
) -> None:
    """Work through the queues of data until all work is done"""

    count_complete_processes = 0
    while count_complete_processes < threads:
        # Always poll exceptions to ensure they're handled
        exception = queues.get_exception()
        if exception is not None:
            handle_exception(
                exception=exception,
                output_handler=output_handler,
                status=status,
                strict=strict,
            )
            continue

        path, data = queues.await_data()

        # Handle exhausted processes
        if path is None:
            # Processed finished sentinel
            count_complete_processes += 1
            logger.info(
                f"Got process end sentinel {count_complete_processes} of {threads}"
            )
            continue

        # Update the progress bar with the total number of converted reads in the file
        if isinstance(data, int):
            status.update_reads_total(path, data)
            output_handler.set_input_complete(path, is_exception=False)
            continue

        # Write the incoming list of converted reads
        writer = output_handler.get_writer(path)
        if writer is None:
            logger.warn(
                f"Trying to write to {path} writer which was closed by an exception"
            )
        else:
            logger.info(f"Writing {len(data)} reads to {path.name} using {writer}")
            writer.add_reads(data)
            status.increment_reads(len(data))

    status.close()


@logged(log_time=True)
def convert_from_fast5(
    inputs: List[Path],
    output: Path,
    recursive: bool = False,
    threads: int = DEFAULT_THREADS,
    one_to_one: Optional[Path] = None,
    force_overwrite: bool = False,
    signal_chunk_size: int = DEFAULT_SIGNAL_CHUNK_SIZE,
    strict: bool = False,
) -> None:
    """
    Convert fast5 files found (optionally recursively) at the given input Paths
    into pod5 file(s). If one_to_one is a Path then the new pod5 files are
    created in a new relative directory structure within output relative to the the
    one_to_one Path.
    """

    if output.is_file() and not force_overwrite:
        raise FileExistsError(
            "Output path points to an existing file and --force-overwrite not set"
        )

    if len(output.parts) > 1:
        output.parent.mkdir(parents=True, exist_ok=True)

    threads = limit_threads(threads)

    pending_fast5s = collect_inputs(inputs, recursive, "*.fast5", threads=threads)
    if not pending_fast5s:
        logger.fatal(f"Found no *.fast5 files in inputs: {inputs}")
        raise RuntimeError("Found no fast5 inputs to process - Exiting")

    output_handler = OutputHandler(output, one_to_one, force_overwrite)
    status = StatusMonitor(pending_fast5s)

    threads = min(threads, len(pending_fast5s))
    ctx = mp.get_context("spawn")
    queues = QueueManager(
        context=ctx,
        inputs=pending_fast5s,
        threads=threads,
        timeout=TIMEOUT_SECONDS,
    )

    active_processes = []
    for _ in range(threads):
        process = ctx.Process(
            target=convert_fast5_files,
            args=(queues, signal_chunk_size),
            daemon=True,
        )
        process.start()
        active_processes.append(process)

    try:
        process_conversion_tasks(
            queues=queues,
            output_handler=output_handler,
            status=status,
            strict=strict,
            threads=threads,
        )

        queues.shutdown()
        for proc in active_processes:
            proc.join()
            proc.close()

    except Exception as exc:
        status.write(f"An unexpected error occurred: {exc}", file=sys.stderr)
        terminate_processes(active_processes)
        raise exc

    finally:
        output_handler.close_all()
        logger.disabled = True


def main():
    """Main function for pod5_convert_from_fast5"""
    run_tool(pod5_convert_from_fast5_argparser())


if __name__ == "__main__":
    main()



"""Main entry point for pod5 tools"""
import argparse
from typing import Any

from pod5 import __version__
from pod5.tools.parsers import (
    SubcommandHelpFormatter,
    prepare_pod5_convert,
    prepare_pod5_inspect_argparser,
    prepare_pod5_merge_argparser,
    prepare_pod5_repack_argparser,
    prepare_pod5_subset_argparser,
    prepare_pod5_filter_argparser,
    prepare_pod5_recover_argparser,
    prepare_pod5_update_argparser,
    prepare_pod5_view_argparser,
    run_tool,
)


def main() -> Any:
    """
    The core pod5 tools function which assembles the argparser and executes the required
    pod5 tool.
    """
    desc = (
        "**********      POD5 Tools      **********\n\n"
        "Tools for inspecting, converting, subsetting and formatting POD5 files"
    )

    parser = argparse.ArgumentParser(
        prog="pod5",
        description=desc,
        epilog="Example: pod5 convert fast5 input.fast5 --output output.pod5",
        formatter_class=SubcommandHelpFormatter,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="Pod5 version: {}".format(__version__),
        help="Show pod5 version and exit.",
    )
    parser.set_defaults(func=lambda **_: parser.print_help())

    root = parser.add_subparsers(title="sub-commands")

    # add sub-parsers to the root argparser
    prepare_pod5_convert(root)
    prepare_pod5_inspect_argparser(root)
    prepare_pod5_merge_argparser(root)
    prepare_pod5_repack_argparser(root)
    prepare_pod5_subset_argparser(root)
    prepare_pod5_filter_argparser(root)
    prepare_pod5_recover_argparser(root)
    prepare_pod5_update_argparser(root)
    prepare_pod5_view_argparser(root)

    # Run the tool
    return run_tool(parser)


if __name__ == "__main__":
    main()



"""
Tools to assist repacking pod5 data into other pod5 files
"""
from typing import Collection
import lib_pod5 as p5b

import pod5 as p5


class Repacker:
    """Wrapper class around native pod5 tools to repack data"""

    def __init__(self):
        self._repacker = p5b.Repacker()
        self._reads_requested = 0

    @property
    def is_complete(self) -> bool:
        """Find if the requested repack operations are complete"""
        return self._repacker.is_complete

    @property
    def currently_open_file_reader_count(self) -> int:
        """Returns the number of open file readers held by this repacker"""
        return self._repacker.currently_open_file_reader_count

    @property
    def reads_completed(self) -> int:
        """Find the number of reads written to files"""
        return self._repacker.reads_completed

    @property
    def reads_requested(self) -> int:
        """Find the number of requested reads to be written"""
        return self._reads_requested

    def add_output(
        self, output_file: p5.Writer, check_duplicate_read_ids: bool = True
    ) -> p5b.Pod5RepackerOutput:
        """
        Add an output file writer to the repacker, so it can have read data repacked
        into it.

        Once a user has added an output, it can be passed as an output
        to :py:meth:`add_selected_reads_to_output` or :py:meth:`add_reads_to_output`

        Parameters
        ----------
        output_file: :py:class:`writer.Writer`
            The output file writer to use
        check_duplicate_read_ids: bool
            Check the output for duplicate read ids, and raise an error if found.

        Returns
        -------
        repacker_object: p5b.Pod5RepackerOutput
            Use this as "output_ref" in calls to :py:meth:`add_selected_reads_to_output`
            or :py:meth:`add_reads_to_output`
        """
        assert output_file._writer is not None
        return self._repacker.add_output(output_file._writer, check_duplicate_read_ids)

    def add_selected_reads_to_output(
        self,
        output_ref: p5b.Pod5RepackerOutput,
        reader: p5.Reader,
        selected_read_ids: Collection[str],
    ):
        """
        Copy the selected read_ids from the given :py:class:`Reader` into the
        Repacker output reference which was returned by :py:meth:`add_output`

        Parameters
        ----------
        output_ref : lib_pod5.pod5_format_pybind.Pod5RepackerOutput
            The repacker handle reference returned from :py:meth:`add_output`
        reader : :py:class:`Reader`
            The Pod5 file reader to copy reads from
        selected_read_ids: Collection[str]
            A Collection of read_ids as strings

        Raises
        ------
        RuntimeError
            If any of the selected_read_ids were not found in the source file
        """

        successful_finds, per_batch_counts, all_batch_rows = reader._plan_traversal(
            selected_read_ids
        )

        if successful_finds != len(selected_read_ids):
            raise RuntimeError(
                f"Failed to find {len(selected_read_ids) - successful_finds} "
                "requested reads in the source file"
            )

        self._reads_requested += successful_finds
        self._repacker.add_selected_reads_to_output(
            output_ref, reader.inner_file_reader, per_batch_counts, all_batch_rows
        )

    def add_all_reads_to_output(
        self, output_ref: p5b.Pod5RepackerOutput, reader: p5.Reader
    ) -> None:
        """
        Copy the every read from the given :py:class:`Reader` into the
        Repacker output reference which was returned by :py:meth:`add_output`

        Parameters
        ----------
        output_ref : lib_pod5.pod5_format_pybind.Pod5RepackerOutput
            The repacker handle reference returned from :py:meth:`add_output`
        reader : :py:class:`Reader`
            The Pod5 file reader to copy reads from
        """
        self._reads_requested += reader.num_reads
        self._repacker.add_all_reads_to_output(output_ref, reader.inner_file_reader)

    def finish(self) -> None:
        """
        Call finish on the underlying c_api repacker instance to write the footer
        completing the file and freeing resources
        """
        return self._repacker.finish()

    def set_output_finished(self, output) -> None:
        """
        Tell the repacker a specific output is complete and can be finalised.
        """
        return self._repacker.set_output_finished(output)
